# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import datetime
import importlib
import os
import pathlib
import re
import sys
import typing

sys.path.insert(0, os.fspath(pathlib.Path.cwd()))
_today = datetime.datetime.now()


# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

package = "mpi4py"


def pkg_version():
    topdir = pathlib.Path(__file__).resolve().parent.parent.parent
    source = topdir / "src" / "mpi4py" / "__init__.py"
    content = source.read_text(encoding="utf-8")
    m = re.search(r'__version__\s*=\s*"(.*)"', content)
    return m.groups()[0]


project = "MPI for Python"
author = "Lisandro Dalcin"
copyright = f"{_today.year}, {author}"  # noqa: A001

release = pkg_version()
version = release.rsplit(".", 1)[0]


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

needs_sphinx = "5.0.0"

default_role = "any"

nitpicky = True
nitpick_ignore = [
    ("c:func", r"atexit"),
    ("c:func", r"dlopen"),
    ("c:func", r"dlsym"),
    ("py:mod", r"__worker__"),
]
nitpick_ignore_regex = [
    (r"c:.*", r"MPI_.*"),
    (r"envvar", r"(LD_LIBRARY_)?PATH"),
    (r"envvar", r"(MPICH|OMPI|MPIEXEC)_.*"),
]

toc_object_entries = False
toc_object_entries_show_parents = "hide"
# python_use_unqualified_type_names = True

autodoc_class_signature = "separated"
autodoc_typehints = "description"
autodoc_typehints_format = "short"
autodoc_mock_imports = []
autodoc_type_aliases = {}

autosummary_context = {
    "synopsis": {},
    "autotype": {},
}

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "dlpack": ("https://dmlc.github.io/dlpack/latest/", None),
    "numba": ("https://numba.readthedocs.io/en/stable/", None),
}

napoleon_preprocess_types = True

try:
    import sphinx_rtd_theme
except ImportError:
    sphinx_rtd_theme = None
else:
    if "sphinx_rtd_theme" not in extensions:
        extensions.append("sphinx_rtd_theme")

try:
    import sphinx_copybutton
except ImportError:
    sphinx_copybutton = None
else:
    if "sphinx_copybutton" not in extensions:
        extensions.append("sphinx_copybutton")

copybutton_exclude = ".linenos, .gp, .go"
copybutton_prompt_text = r"\$ |>>> |\.\.\. "
copybutton_prompt_is_regexp = True
copybutton_line_continuation_character = "\\"

extensions.append("sphinx.ext.coverage")
coverage_ignore_modules = [r"mpi4py\.(bench|run)"]
coverage_ignore_classes = [r"Rc", r"memory"]


def _setup_numpy_typing():
    try:
        import numpy as np
    except ImportError:
        from types import new_class
        from typing import Generic, TypeVar

        np = type(sys)("numpy")
        sys.modules[np.__name__] = np
        np.dtype = new_class("dtype", (Generic[TypeVar("T")],))
        np.dtype.__module__ = np.__name__

    try:
        import numpy.typing as npt
    except ImportError:
        npt = type(sys)("numpy.typing")
        np.typing = npt
        sys.modules[npt.__name__] = npt
        npt.__all__ = []
        for attr in ["ArrayLike", "DTypeLike"]:
            setattr(npt, attr, typing.Any)
            npt.__all__.append(attr)


def _patch_domain_python():
    try:
        from numpy.typing import __all__ as numpy_types
    except ImportError:
        numpy_types = []
    try:
        from mpi4py.typing import __all__ as mpi4py_types
    except ImportError:
        mpi4py_types = []

    numpy_types = set(numpy_types)
    mpi4py_types = set(mpi4py_types)
    for name in numpy_types:
        autodoc_type_aliases[name] = f"~numpy.typing.{name}"
    for name in mpi4py_types:
        autodoc_type_aliases[name] = f"~mpi4py.typing.{name}"

    from sphinx.domains.python import PythonDomain
    from sphinx.util.inspect import TypeAliasForwardRef

    PythonDomain.object_types["data"].roles += ("class",)
    TypeAliasForwardRef.__repr__ = lambda self: self.name


def _setup_autodoc(app):
    from sphinx.ext import autodoc, autosummary
    from sphinx.locale import _
    from sphinx.util.typing import restify

    #

    class ClassDocumenterMixin:
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            if self.config.autodoc_class_signature == "separated":
                members = self.options.members
                special_members = self.options.special_members
                if special_members is not None:
                    for name in ("__new__", "__init__"):
                        if name in members:
                            members.remove(name)
                        if name in special_members:
                            special_members.remove(name)

    class ClassDocumenter(
        ClassDocumenterMixin,
        autodoc.ClassDocumenter,
    ):
        pass

    class ExceptionDocumenter(
        ClassDocumenterMixin,
        autodoc.ExceptionDocumenter,
    ):
        pass

    app.add_autodocumenter(ClassDocumenter, override=True)
    app.add_autodocumenter(ExceptionDocumenter, override=True)

    #

    def istypealias(obj, name):
        if isinstance(obj, type):
            return name != getattr(obj, "__name__", None)
        return obj == typing.Any

    def istypevar(obj):
        return isinstance(obj, typing.TypeVar)

    class TypeDocumenter(autodoc.DataDocumenter):
        objtype = "type"
        directivetype = "data"
        priority = autodoc.ClassDocumenter.priority + 1

        @classmethod
        def can_document_member(cls, member, membername, _isattr, parent):
            return (
                isinstance(parent, autodoc.ModuleDocumenter)
                and parent.name == "mpi4py.typing"
                and (istypevar(member) or istypealias(member, membername))
            )

        def add_directive_header(self, sig):
            if istypevar(self.object):
                obj = self.object
                if not self.options.annotation:
                    self.options.annotation = f' = TypeVar("{obj.__name__}")'
            super().add_directive_header(sig)

        def update_content(self, more_content):
            obj = self.object
            if istypevar(obj):
                if obj.__covariant__:
                    kind = _("Covariant")
                elif obj.__contravariant__:
                    kind = _("Contravariant")
                else:
                    kind = _("Invariant")
                content = f"{kind} :class:`~typing.TypeVar`."
                more_content.append(content, "")
                more_content.append("", "")
            if istypealias(obj, self.name):
                content = _("alias of %s") % restify(obj)
                more_content.append(content, "")
                more_content.append("", "")
            super().update_content(more_content)

        def get_doc(self, *args, **kwargs):
            obj = self.object
            if istypevar(obj):
                if obj.__doc__ == typing.TypeVar.__doc__:
                    return []
            return super().get_doc(*args, **kwargs)

    app.add_autodocumenter(TypeDocumenter)

    #

    class ExceptionDocumenterCustom(ExceptionDocumenter):
        objtype = "class"

    def get_documenter(*args, **kwargs):
        if hasattr(autosummary, "_get_documenter"):
            obj = args[0]  # signature is (obj, parent, *, registry)
            get_documenter = autosummary._get_documenter
        else:
            obj = args[1]  # signature is (app, obj, parent)
            get_documenter = autosummary.get_documenter
        if isinstance(obj, type) and issubclass(obj, BaseException):
            caller = sys._getframe().f_back.f_code.co_name
            if caller == "generate_autosummary_content":
                if obj.__module__ == "mpi4py.MPI":
                    if obj.__name__ == "Exception":
                        return ExceptionDocumenterCustom
        return get_documenter(*args, **kwargs)

    from sphinx.ext.autosummary import generate

    if hasattr(generate, "_get_documenter"):
        generate._get_documenter = get_documenter
    else:
        generate.get_documenter = get_documenter


def setup(app):
    _setup_numpy_typing()
    _patch_domain_python()
    _setup_autodoc(app)

    try:
        from mpi4py import MPI
    except ImportError:
        autodoc_mock_imports.append("mpi4py")
        return

    sys_dwb = sys.dont_write_bytecode
    sys.dont_write_bytecode = True
    apidoc = __import__("apidoc")
    sys.dont_write_bytecode = sys_dwb

    name = MPI.__name__
    here = pathlib.Path(__file__).resolve().parent
    outdir = here / apidoc.OUTDIR
    source = outdir / f"{name}.py"
    getmtime = os.path.getmtime
    generate = (
        not source.exists()
        or getmtime(source) < getmtime(MPI.__file__)
        or getmtime(source) < getmtime(apidoc.__file__)
    )
    if generate:
        apidoc.generate(source)
    module = apidoc.load_module(source)
    module.Get_version.__code__ = (lambda: (4, 0)).__code__
    apidoc.replace_module(module)

    synopsis = autosummary_context["synopsis"]
    synopsis[module.__name__] = module.__doc__.strip()
    autotype = autosummary_context["autotype"]
    autotype[module.Exception.__name__] = "exception"

    modules = [
        "mpi4py",
        "mpi4py.run",
        "mpi4py.util.dtlib",
        "mpi4py.util.pkl5",
        "mpi4py.util.pool",
        "mpi4py.util.sync",
    ]
    typing_overload = typing.overload
    typing.overload = lambda arg: arg
    for name in modules:
        mod = importlib.import_module(name)
        ann = apidoc.load_module(f"{mod.__file__}i", name)
        apidoc.annotate(mod, ann)
    typing.overload = typing_overload


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = (
    "sphinx_rtd_theme" if "sphinx_rtd_theme" in extensions else "default"
)

html_logo = "../mpi4py.svg"

html_favicon = "../mpi4py.svg"

if html_theme == "default":
    html_copy_source = False


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = f"{package}-man"


# -- Options for LaTeX output ---------------------------------------------

# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    ("index", f"{package}.tex", project, author, "howto"),
]

latex_elements = {
    "papersize": "a4",
}


# -- Options for manual page output ---------------------------------------

# (source start file, name, description, authors, manual section).
man_pages = [("index", package, project, [author], 3)]


# -- Options for Texinfo output -------------------------------------------

# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        "index",
        package,
        project,
        author,
        package,
        f"{project}.",
        "Miscellaneous",
    ),
]


# -- Options for Epub output ----------------------------------------------

# Output file base name for ePub builder.
epub_basename = package



# mpiexec -n 3 python ex-2.29.py

# Use a blocking probe to wait for an incoming message

# --------------------------------------------------------------------

import array

from mpi4py import MPI

if MPI.COMM_WORLD.Get_size() < 3:
    raise SystemExit from None

# --------------------------------------------------------------------

comm = MPI.COMM_WORLD
rank = comm.Get_rank()

if rank == 0:
    i = array.array("i", [7] * 5)
    comm.Send([i, MPI.INT], 2, 0)
elif rank == 1:
    x = array.array("f", [7] * 5)
    comm.Send([x, MPI.FLOAT], 2, 0)
elif rank == 2:
    i = array.array("i", [0] * 5)
    x = array.array("f", [0] * 5)
    status = MPI.Status()
    for _j in range(2):
        comm.Probe(MPI.ANY_SOURCE, 0, status)
        if status.Get_source() == 0:
            comm.Recv([i, MPI.INT], 0, 0, status)
        else:
            comm.Recv([x, MPI.FLOAT], 1, 0, status)

# --------------------------------------------------------------------

if rank == 2:
    for v in i:
        assert v == 7
    for v in x:
        assert v == 7
    assert status.source in (0, 1)
    assert status.tag == 0
    assert status.error == 0

# --------------------------------------------------------------------



# Author:  Lisandro Dalcin
# Contact: dalcinl@gmail.com
import argparse
import fnmatch
import importlib
import os
import pathlib
import re
import sys
import unittest

__unittest = True


def parse_pattern(pattern):
    return f"*{pattern}*" if "*" not in pattern else pattern


def parse_xfile(xfile):
    content = pathlib.Path(xfile).read_text(encoding="utf-8")
    return list(map(parse_pattern, content.splitlines()))


def setup_parser(parser):
    parser.add_argument(
        "-x",  # "--exclude-name",
        help="Skip run tests which match the given substring",
        action="append",
        dest="excludePatterns",
        default=[],
        type=parse_pattern,
        metavar="TESTNAMEPATTERNS",
    )
    parser.add_argument(  # Py3.8: use action="extend" on "excludePatterns"
        "--xfile",
        help="Skip run tests which match the substrings in the given files",
        action="append",
        dest="excludeFile",
        default=[],
        type=str,
        metavar="FILENAME",
    )
    parser.add_argument(
        "-i",  # "--include",
        help="Include test module names matching REGEX",
        action="append",
        dest="include",
        default=[],
        type=str,
        metavar="REGEX",
    )
    parser.add_argument(
        "-e",  # "--exclude",
        help="Exclude test module names matching REGEX",
        action="append",
        dest="exclude",
        default=[],
        type=str,
        metavar="REGEX",
    )
    parser.add_argument(
        "--inplace",
        help="Enable testing from in-place build",
        action="store_true",
        dest="inplace",
        default=False,
    )
    parser.add_argument(
        "--no-builddir",
        help="Disable testing from build directory",
        action="store_false",
        dest="builddir",
        default=True,
    )
    parser.add_argument(
        "--path",
        help="Prepend PATH to sys.path",
        action="append",
        dest="path",
        default=[],
        type=str,
        metavar="PATH",
    )
    parser.add_argument(
        "--threads",
        help="Initialize MPI with thread support",
        action="store_true",
        dest="threads",
        default=None,
    )
    parser.add_argument(
        "--no-threads",
        help="Initialize MPI without thread support",
        action="store_false",
        dest="threads",
        default=None,
    )
    parser.add_argument(
        "--thread-level",
        help="Initialize MPI with required thread support",
        choices=["single", "funneled", "serialized", "multiple"],
        action="store",
        dest="thread_level",
        default=None,
        type=str,
        metavar="LEVEL",
    )
    parser.add_argument(
        "--cupy",
        help="Enable testing with CuPy arrays",
        action="store_true",
        dest="cupy",
        default=False,
    )
    parser.add_argument(
        "--no-cupy",
        help="Disable testing with CuPy arrays",
        action="store_false",
        dest="cupy",
        default=False,
    )
    parser.add_argument(
        "--numba",
        help="Enable testing with Numba arrays",
        action="store_true",
        dest="numba",
        default=False,
    )
    parser.add_argument(
        "--no-numba",
        help="Disable testing with Numba arrays",
        action="store_false",
        dest="numba",
        default=False,
    )
    parser.add_argument(
        "--no-numpy",
        help="Disable testing with NumPy arrays",
        action="store_false",
        dest="numpy",
        default=True,
    )
    parser.add_argument(
        "--no-array",
        help="Disable testing with builtin array module",
        action="store_false",
        dest="array",
        default=True,
    )
    parser.add_argument(
        "--no-skip-mpi",
        help="Disable known failures with backend MPI",
        action="store_false",
        dest="skip_mpi",
        default=True,
    )
    parser.add_argument(
        "--xml",
        help="Directory for storing XML reports",
        action="store",
        dest="xmloutdir",
        default=None,
        nargs="?",
        const=os.path.curdir,
    )
    return parser


def getbuilddir():
    try:
        try:
            from setuptools.dist import Distribution
        except ImportError:
            from distutils.dist import Distribution
        try:
            from setuptools.command.build import build
        except ImportError:
            from distutils.command.build import build
        cmd_obj = build(Distribution())
        cmd_obj.finalize_options()
        builddir = pathlib.Path(cmd_obj.build_platlib)
    except Exception:
        builddir = None
    return builddir


def getprocessorinfo():
    MPI = importlib.import_module("mpi4py.MPI")
    rank = MPI.COMM_WORLD.Get_rank()
    name = MPI.Get_processor_name()
    return (rank, name)


def getlibraryinfo():
    MPI = importlib.import_module("mpi4py.MPI")
    x, y = MPI.Get_version()
    info = f"MPI {x}.{y}"
    name, version = MPI.get_vendor()
    if name != "unknown":
        x, y, z = version
        info += f" ({name} {x}.{y}.{z})"
    return info


def getpythoninfo():
    x, y, z = sys.version_info[:3]
    return f"Python {x}.{y}.{z} ({sys.executable})"


def getpackageinfo(pkg):
    try:
        pkg = __import__(pkg)
    except ImportError:
        return None
    name = pkg.__name__
    version = pkg.__version__
    path = pkg.__path__[0]
    return f"{name} {version} ({path})"


def setup_python(options):
    script = pathlib.Path(__file__).resolve()
    testdir = script.parent
    rootdir = testdir.parent
    if options.inplace:
        srcdir = rootdir / "src"
        sys.path.insert(0, os.fspath(srcdir))
    elif options.builddir:
        builddir = getbuilddir()
        if builddir is not None:
            builddir = rootdir / builddir
            if builddir.is_dir():
                sys.path.insert(0, os.fspath(builddir))
    if options.path:
        for path in reversed(options.path):
            for pth in map(pathlib.Path, path.split(os.path.pathsep)):
                if pth.exists():
                    sys.path.insert(0, os.fspath(pth))


def setup_modules(options):
    #
    if not options.cupy:
        sys.modules["cupy"] = None
    if not options.numba:
        sys.modules["numba"] = None
    if not options.numpy:
        sys.modules["numpy"] = None
    if not options.array:
        sys.modules["array"] = None
    #
    mpi4py = importlib.import_module("mpi4py")
    if options.threads is not None:
        mpi4py.rc.threads = options.threads
    if options.thread_level is not None:
        mpi4py.rc.thread_level = options.thread_level
    importlib.import_module("mpi4py.MPI")


def setup_unittest(_options):
    from contextlib import suppress
    from unittest.runner import _WritelnDecorator

    super_writeln = _WritelnDecorator.writeln

    def writeln(self, arg=None):
        with suppress(Exception):
            self.flush()
        super_writeln(self, arg)
        with suppress(Exception):
            self.flush()

    _WritelnDecorator.writeln = writeln


def print_banner(options):
    rank, name = getprocessorinfo()
    prefix = f"[{rank}@{name}]"

    def writeln(message="", endl="\n"):
        if message is None:
            return
        sys.stderr.flush()
        sys.stderr.write(f"{prefix} {message}{endl}")
        sys.stderr.flush()

    if options.verbosity:
        writeln(getpythoninfo())
        writeln(getpackageinfo("numpy"))
        writeln(getpackageinfo("numba"))
        writeln(getpackageinfo("cupy"))
        writeln(getlibraryinfo())
        writeln(getpackageinfo("mpi4py"))


class TestLoader(unittest.TestLoader):
    #
    excludePatterns = None

    def __init__(self, include=None, exclude=None):
        super().__init__()
        if include:
            self.include = re.compile("|".join(include)).search
        else:
            self.include = lambda _arg: True
        if exclude:
            self.exclude = re.compile("|".join(exclude)).search
        else:
            self.exclude = lambda _arg: False

    def _match_path(self, path, full_path, pattern):
        match = super()._match_path(path, full_path, pattern)
        if match:
            if not self.include(path):
                return False
            if self.exclude(path):
                return False
        return match

    def getTestCaseNames(self, testCaseClass):
        def exclude(name):
            modname = testCaseClass.__module__
            clsname = testCaseClass.__qualname__
            fullname = f"{modname}.{clsname}.{name}"
            return not any(
                fnmatch.fnmatchcase(fullname, pattern)
                for pattern in self.excludePatterns
            )

        names = super().getTestCaseNames(testCaseClass)
        if self.excludePatterns:
            names = list(filter(exclude, names))
        return names


class TestProgram(unittest.TestProgram):
    #
    def _getMainArgParser(self, parent):
        parser = super()._getMainArgParser(parent)
        setup_parser(parser)
        return parser

    def _getDiscoveryArgParser(self, parent):
        parser = argparse.ArgumentParser(parents=[parent])
        parser.color = True  # Python 3.14+
        setup_parser(parser)
        return parser

    def createTests(self, from_discovery=False, Loader=None):
        setup_python(self)
        setup_modules(self)
        setup_unittest(self)
        testdir = os.fspath(pathlib.Path(__file__).resolve().parent)
        if from_discovery:
            self.start = testdir
            self.pattern = "test_*.py"
        elif testdir not in sys.path:
            sys.path.insert(0, os.fspath(testdir))
        if not self.skip_mpi:
            mpiunittest = __import__("mpiunittest")
            mpiunittest.skipMPI = lambda _p, *_c: lambda f: f
        for xfile in self.excludeFile:
            self.excludePatterns.extend(parse_xfile(xfile))
        self.testLoader = TestLoader(self.include, self.exclude)
        self.testLoader.excludePatterns = self.excludePatterns
        super().createTests(from_discovery, Loader)

    def _setUpXMLRunner(self):
        MPI = importlib.import_module("mpi4py.MPI")
        size = MPI.COMM_WORLD.Get_size()
        rank = MPI.COMM_WORLD.Get_rank()
        try:
            import xmlrunner
        except ModuleNotFoundError:
            if rank == 0:
                print(
                    "Cannot generate XML reports!",
                    "Install 'unittest-xml-reporting'.",
                    file=sys.stderr,
                    flush=True,
                )
            sys.exit(1)
        runner = xmlrunner.XMLTestRunner(output=self.xmloutdir)
        runner.outsuffix += f"-{rank}" if size > 1 else ""
        self.testRunner = runner
        self.buffer = False
        self.catchbreak = False

    def runTests(self):
        print_banner(self)
        if self.xmloutdir:
            self._setUpXMLRunner()
        try:
            super().runTests()
        except SystemExit:
            pass
        success = self.result.wasSuccessful()
        if not success and self.failfast:
            run = importlib.import_module("mpi4py.run")
            run.set_abort_status(1)
        sys.exit(not success)


main = TestProgram


if __name__ == "__main__":
    sys.dont_write_bytecode = True
    main(module=None)



# Author:  Lisandro Dalcin
# Contact: dalcinl@gmail.com
"""Run MPI benchmarks and tests."""

import sys as _sys

_registry = {}


def _register(command):
    _registry[command.__name__] = command
    return command


def _create_parser(cmd="", usage=None):
    # pylint: disable=import-outside-toplevel
    from argparse import ArgumentParser

    pyexe = __import__("pathlib").Path(_sys.executable).name
    pycmd = f"{pyexe} -m {__spec__.name} {cmd}".strip()
    parser = ArgumentParser(prog=pycmd, usage=usage)
    parser.color = True  # Python 3.14+
    return parser


@_register
def helloworld(comm, args=None, verbose=True):
    """Hello, World! using MPI."""
    # pylint: disable=import-outside-toplevel

    parser = _create_parser("helloworld")
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_false",
        dest="verbose",
        default=verbose,
        help="quiet output",
    )
    options = parser.parse_args(args)

    from . import MPI

    size = comm.Get_size()
    rank = comm.Get_rank()
    name = MPI.Get_processor_name()
    message = (
        f"Hello, World! I am process "
        f"{rank:{len(str(size - 1))}d} "
        f"of {size} on {name}.\n"
    )
    comm.Barrier()
    if rank > 0:
        comm.Recv([None, "B"], rank - 1)
    if options.verbose:
        _sys.stdout.write(message)
        _sys.stdout.flush()
    if rank < size - 1:
        comm.Send([None, "B"], rank + 1)
    comm.Barrier()
    return message


@_register
def ringtest(comm, args=None, verbose=True):
    """Time a message going around the ring of processes."""
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-statements
    # pylint: disable=import-outside-toplevel

    parser = _create_parser("ringtest")
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_false",
        dest="verbose",
        default=verbose,
        help="quiet output",
    )
    parser.add_argument(
        "-n",
        "--size",
        type=int,
        dest="size",
        default=1,
        help="message size",
    )
    parser.add_argument(
        "-s",
        "--skip",
        type=int,
        dest="skip",
        default=0,
        help="number of warm-up iterations",
    )
    parser.add_argument(
        "-l",
        "--loop",
        type=int,
        dest="loop",
        default=1,
        help="number of iterations",
    )
    options = parser.parse_args(args)

    def ring(comm, n=1, loop=1, skip=0):
        # pylint: disable=invalid-name
        from array import array

        from . import MPI

        iterations = list(range(loop + skip))
        size = comm.Get_size()
        rank = comm.Get_rank()
        source = (rank - 1) % size
        dest = (rank + 1) % size
        Sendrecv = comm.Sendrecv
        Send = comm.Send
        Recv = comm.Recv
        Wtime = MPI.Wtime
        sendmsg = array("B", [+42]) * n
        recvmsg = array("B", [0x0]) * n
        tic = Wtime()
        if size == 1:
            for i in iterations:
                if i == skip:
                    tic = Wtime()
                Sendrecv(sendmsg, dest, 0, recvmsg, source, 0)
        else:
            if rank == 0:
                for i in iterations:
                    if i == skip:
                        tic = Wtime()
                    Send(sendmsg, dest, 0)
                    Recv(recvmsg, source, 0)
            else:
                sendmsg = recvmsg
                for i in iterations:
                    if i == skip:
                        tic = Wtime()
                    Recv(recvmsg, source, 0)
                    Send(sendmsg, dest, 0)
        toc = Wtime()
        if comm.rank == 0 and sendmsg != recvmsg:  # pragma: no cover
            import traceback
            from warnings import warn

            try:
                warn("received message does not match!", stacklevel=1)
            except UserWarning:
                traceback.print_exc()
                comm.Abort(2)
        return toc - tic

    size = getattr(options, "size", 1)
    loop = getattr(options, "loop", 1)
    skip = getattr(options, "skip", 0)
    comm.Barrier()
    elapsed = ring(comm, size, loop, skip)
    if options.verbose and comm.rank == 0:
        _sys.stdout.write(
            f"time for {loop} loops = {elapsed:g} seconds "
            f"({comm.size} processes, {size} bytes)\n"
        )
        _sys.stdout.flush()
    return elapsed


@_register
def pingpong(comm, args=None, verbose=True):
    """Time messages between processes."""
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-statements
    # pylint: disable=import-outside-toplevel

    parser = _create_parser("pingpong")
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_false",
        dest="verbose",
        default=verbose,
        help="quiet output",
    )
    parser.add_argument(
        "-m",
        "--min-size",
        type=int,
        dest="min_size",
        default=1,
        help="minimum message size",
    )
    parser.add_argument(
        "-n",
        "--max-size",
        type=int,
        dest="max_size",
        default=1 << 30,
        help="maximum message size",
    )
    parser.add_argument(
        "-s",
        "--skip",
        type=int,
        dest="skip",
        default=100,
        help="number of warm-up iterations",
    )
    parser.add_argument(
        "-l",
        "--loop",
        type=int,
        dest="loop",
        default=10000,
        help="number of iterations",
    )
    parser.add_argument(
        "-a",
        "--array",
        action="store",
        dest="array",
        default="numpy",
        choices=["numpy", "cupy", "numba", "none"],
        help="use NumPy/CuPy/Numba arrays",
    )
    parser.add_argument(
        "-p",
        "--pickle",
        action="store_true",
        dest="pickle",
        default=False,
        help="use pickle-based send and receive",
    )
    parser.add_argument(
        "--protocol",
        type=int,
        dest="protocol",
        default=None,
        help="pickle protocol version",
    )
    parser.add_argument(
        "-o",
        "--outband",
        action="store_true",
        dest="outband",
        default=False,
        help="use out-of-band pickle-based send and receive",
    )
    parser.add_argument(
        "--threshold",
        type=int,
        dest="threshold",
        default=None,
        help="size threshold for out-of-band pickle buffers",
    )
    parser.add_argument(
        "--skip-large",
        type=int,
        dest="skip_large",
        default=10,
    )
    parser.add_argument(
        "--loop-large",
        type=int,
        dest="loop_large",
        default=1000,
    )
    parser.add_argument(
        "--large-size",
        type=int,
        dest="large_size",
        default=1 << 14,
    )
    parser.add_argument(
        "--skip-huge",
        type=int,
        dest="skip_huge",
        default=1,
    )
    parser.add_argument(
        "--loop-huge",
        type=int,
        dest="loop_huge",
        default=10,
    )
    parser.add_argument(
        "--huge-size",
        type=int,
        dest="huge_size",
        default=1 << 20,
    )
    parser.add_argument(
        "--no-header",
        action="store_false",
        dest="print_header",
        default=True,
    )
    parser.add_argument(
        "--no-stats",
        action="store_false",
        dest="print_stats",
        default=True,
    )
    options = parser.parse_args(args)

    import statistics

    from . import MPI
    from .util import pkl5

    # pylint: disable=import-error
    numpy = cupy = numba = None
    if options.array == "numpy":
        try:
            import numpy
        except ImportError:  # pragma: no cover
            pass
    elif options.array == "cupy":  # pragma: no cover
        import cupy
    elif options.array == "numba":  # pragma: no cover
        import numba.cuda

    skip = options.skip
    loop = options.loop
    min_size = options.min_size
    max_size = options.max_size
    skip_large = options.skip_large
    loop_large = options.loop_large
    large_size = options.large_size
    skip_huge = options.skip_huge
    loop_huge = options.loop_huge
    huge_size = options.huge_size

    use_pickle = options.pickle or options.outband
    use_outband = options.outband
    protocol = options.protocol if use_pickle else None
    threshold = options.threshold if use_outband else None

    if use_outband:
        comm = pkl5.Intracomm(comm)
    if protocol is not None:
        MPI.pickle.PROTOCOL = protocol
    if threshold is not None:
        pkl5.pickle.THRESHOLD = threshold

    buf_sizes = [1 << i for i in range(33)]
    buf_sizes = [n for n in buf_sizes if min_size <= n <= max_size]

    wtime = MPI.Wtime
    if use_pickle:
        send = comm.send
        recv = comm.recv
        sendrecv = comm.sendrecv
    else:
        send = comm.Send
        recv = comm.Recv
        sendrecv = comm.Sendrecv
    s_msg = r_msg = None

    def allocate(nbytes):  # pragma: no cover
        if numpy:
            return numpy.empty(nbytes, "B")
        elif cupy:
            return cupy.empty(nbytes, "B")
        elif numba:
            return numba.cuda.device_array(nbytes, "B")
        else:
            return bytearray(nbytes)

    def run_pingpong():
        rank = comm.Get_rank()
        size = comm.Get_size()
        t_start = wtime()
        if size == 1:
            sendrecv(s_msg, 0, 0, r_msg, 0, 0)
            sendrecv(s_msg, 0, 0, r_msg, 0, 0)
        elif rank == 0:
            send(s_msg, 1, 0)
            recv(r_msg, 1, 0)
        elif rank == 1:
            recv(r_msg, 0, 0)
            send(s_msg, 0, 0)
        t_end = wtime()
        return (t_end - t_start) / 2

    result = []
    for nbytes in buf_sizes:
        if nbytes > large_size:
            skip = min(skip, skip_large)
            loop = min(loop, loop_large)
        if nbytes > huge_size:
            skip = min(skip, skip_huge)
            loop = min(loop, loop_huge)
        iterations = list(range(loop + skip))

        if use_pickle:
            s_msg = allocate(nbytes)
        else:
            s_msg = [allocate(nbytes), nbytes, MPI.BYTE]
            r_msg = [allocate(nbytes), nbytes, MPI.BYTE]

        t_list = []
        comm.Barrier()
        for i in iterations:
            elapsed = run_pingpong()
            if i >= skip:
                t_list.append(elapsed)

        s_msg = r_msg = None

        t_mean = statistics.mean(t_list) if t_list else float("nan")
        t_stdev = statistics.stdev(t_list) if len(t_list) > 1 else 0.0
        result.append((nbytes, t_mean, t_stdev))

        if options.verbose and comm.rank == 0:
            if options.print_header:
                options.print_header = False
                print("# MPI PingPong Test")
                header = "# Size [B]  Bandwidth [MB/s]"
                if options.print_stats:
                    header += " | Time Mean [s] \u00b1 StdDev [s]  Samples"
                print(header, flush=True)
            bandwidth = nbytes / t_mean
            message = f"{nbytes:10d}{bandwidth / 1e6:18.2f}"
            if options.print_stats:
                message += f" | {t_mean:.7e} \u00b1 {t_stdev:.4e} {loop:8d}"
            print(message, flush=True)

    return result


def _fn_identity(arg):  # pragma: no cover
    return arg


@_register
def futures(comm, args=None, verbose=True):
    """Measure mpi4py.futures task throughput."""
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-statements
    # pylint: disable=import-outside-toplevel

    parser = _create_parser("futures")
    parser.add_argument(
        "-q",
        "--quiet",
        help="quiet output",
        action="store_false",
        dest="verbose",
        default=verbose,
    )
    parser.add_argument(
        "-e",
        "--executor",
        help="executor backend",
        action="store",
        dest="executor",
        default="mpi",
        choices=["mpi", "process", "thread", "interpreter"],
    )
    parser.add_argument(
        "-w",
        "--num-workers",
        help="number or workers",
        type=int,
        dest="workers",
        default=None,
    )
    parser.add_argument(
        "-t",
        "--num-tasks",
        help="number of tasks per worker",
        type=int,
        dest="tasks",
        default=50,
    )
    parser.add_argument(
        "-m",
        "--min-size",
        help="minimum task data size",
        type=int,
        dest="min_size",
        default=0,
    )
    parser.add_argument(
        "-n",
        "--max-size",
        help="maximum task data size",
        type=int,
        dest="max_size",
        default=1 << 20,
    )
    parser.add_argument(
        "-a",
        "--allocator",
        help="task data allocator",
        action="store",
        dest="allocator",
        default="numpy",
        choices=["numpy", "array", "bytes"],
    )
    parser.add_argument(
        "-c",
        "--chunksize",
        help="chunksize parameter",
        type=int,
        dest="chunksize",
        default=1,
    )
    parser.add_argument(
        "-b",
        "--backoff",
        help="backoff parameter",
        type=float,
        dest="backoff",
        default=0.0,
    )
    parser.add_argument(
        "-o",
        "--outband",
        help="use out-of-band pickle",
        action="store_true",
        dest="outband",
        default=False,
    )
    parser.add_argument(
        "-s",
        "--skip",
        help="number of warm-up iterations",
        type=int,
        dest="skip",
        default=1,
    )
    parser.add_argument(
        "-l",
        "--loop",
        help="number of sample iterations",
        type=int,
        dest="loop",
        default=10,
    )
    parser.add_argument(
        "--skip-large",
        type=int,
        dest="skip_large",
        default=1,
    )
    parser.add_argument(
        "--loop-large",
        type=int,
        dest="loop_large",
        default=5,
    )
    parser.add_argument(
        "--large-size",
        type=int,
        dest="large_size",
        default=1 << 16,
    )
    parser.add_argument(
        "--skip-huge",
        type=int,
        dest="skip_huge",
        default=1,
    )
    parser.add_argument(
        "--loop-huge",
        type=int,
        dest="loop_huge",
        default=3,
    )
    parser.add_argument(
        "--huge-size",
        type=int,
        dest="huge_size",
        default=1 << 18,
    )
    parser.add_argument(
        "--no-header",
        action="store_false",
        dest="print_header",
        default=True,
    )
    parser.add_argument(
        "--no-stats",
        action="store_false",
        dest="print_stats",
        default=True,
    )
    options = parser.parse_args(args)

    import concurrent.futures
    import statistics
    import time

    from .futures import MPIPoolExecutor

    executor_type = options.executor
    workers = options.workers
    tasks = options.tasks
    allocator = options.allocator
    backoff = options.backoff
    use_pkl5 = options.outband
    chunksize = options.chunksize

    skip = options.skip
    loop = options.loop
    min_size = options.min_size
    max_size = options.max_size
    skip_large = options.skip_large
    loop_large = options.loop_large
    large_size = options.large_size
    skip_huge = options.skip_huge
    loop_huge = options.loop_huge
    huge_size = options.huge_size

    buf_sizes = [1 << i for i in range(33)]
    buf_sizes = [n for n in buf_sizes if min_size <= n <= max_size]

    wtime = time.perf_counter

    numpy = array = None
    if allocator == "numpy":
        try:
            import numpy
        except ImportError:  # pragma: no cover
            pass
    elif allocator == "array":
        import array

    def allocate(nbytes):
        if numpy:
            return numpy.zeros(nbytes, "B")
        if array:
            buf = array.array("B", [])
            buf.frombytes(bytes(nbytes))
            return buf
        return bytes(nbytes)

    def create_executor():
        if executor_type == "process":
            return concurrent.futures.ProcessPoolExecutor(
                max_workers=workers,
            )
        if executor_type == "thread":
            return concurrent.futures.ThreadPoolExecutor(
                max_workers=workers,
            )
        if executor_type == "interpreter":  # pragma: no cover
            return concurrent.futures.InterpreterPoolExecutor(
                max_workers=workers,
            )
        assert executor_type == "mpi"  # noqa: S101
        return MPIPoolExecutor(
            max_workers=workers,
            backoff=backoff,
            use_pkl5=use_pkl5,
        )

    def get_num_workers():
        return executor._max_workers  # pylint: disable=protected-access

    def prime_executor():
        executor_map(time.sleep, [0.001] * get_num_workers())

    def executor_map(task, data):
        iterator = executor.map(
            task,
            data,
            chunksize=chunksize,
        )
        for _ in iterator:
            pass

    def run_futures():
        t_start = wtime()
        executor_map(_fn_identity, data)
        t_end = wtime()
        return t_end - t_start

    executor = create_executor()
    num_workers = get_num_workers()
    num_tasks = num_workers * tasks

    result = []
    prime_executor()
    for nbytes in buf_sizes:
        if nbytes > large_size:
            skip = min(skip, skip_large)
            loop = min(loop, loop_large)
        if nbytes > huge_size:
            skip = min(skip, skip_huge)
            loop = min(loop, loop_huge)
        iterations = list(range(loop + skip))

        data = [allocate(nbytes) for _ in range(num_tasks)]

        t_list = []
        for i in iterations:
            elapsed = run_futures()
            if i >= skip:
                t_list.append(elapsed)

        data = None

        t_mean = statistics.mean(t_list) if t_list else float("nan")
        t_stdev = statistics.stdev(t_list) if len(t_list) > 1 else 0.0
        result.append((nbytes, t_mean, t_stdev))

        if options.verbose and comm.rank == 0:
            if options.print_header:
                options.print_header = False
                print(
                    f"# {type(executor).__name__} - "
                    f"{num_workers} workers, "
                    f"{tasks} tasks/worker"
                )
                header = "# Size [B]  Tasks/s"
                if options.print_stats:
                    header += " | Time Mean [s] \u00b1 StdDev [s]  Samples"
                print(header, flush=True)
            throughput = num_tasks / t_mean
            message = f"{nbytes:10d}{throughput:9.0f}"
            if options.print_stats:
                message += f" | {t_mean:.7e} \u00b1 {t_stdev:.4e} {loop:8d}"
            print(message, flush=True)

    executor.shutdown()
    return result


def main(args=None):
    """Entry-point for ``python -m mpi4py.bench``."""
    # pylint: disable=import-outside-toplevel
    from argparse import REMAINDER

    parser = _create_parser(usage="%(prog)s [options] <command> [args]")
    parser.add_argument(
        "--threads",
        action="store_true",
        dest="threads",
        default=None,
        help="initialize MPI with thread support",
    )
    parser.add_argument(
        "--no-threads",
        action="store_false",
        dest="threads",
        default=None,
        help="initialize MPI without thread support",
    )
    parser.add_argument(
        "--thread-level",
        dest="thread_level",
        default=None,
        action="store",
        metavar="LEVEL",
        choices="single funneled serialized multiple".split(),
        help="initialize MPI with required thread level",
    )
    parser.add_argument(
        "command",
        action="store",
        metavar="<command>",
        help="benchmark command to run",
    )
    parser.add_argument(
        "args",
        nargs=REMAINDER,
        metavar="[args]",
        help="arguments for benchmark command",
    )
    options = parser.parse_args(args)

    from . import rc

    if options.threads is not None:
        rc.threads = options.threads
    if options.thread_level is not None:
        rc.thread_level = options.thread_level

    from . import MPI

    comm = MPI.COMM_WORLD
    command = _registry.get(options.command)
    if command is None:
        if comm.rank == 0:
            parser.error(f"unknown command {options.command!r}")
        parser.exit(2)
    command(comm, options.args)
    parser.exit()


if __name__ == "__main__":
    main()



import mpiunittest as unittest

from mpi4py import MPI


class BaseTestHandle:
    #
    HANDLES = []

    def checkHandleMPI(self, handle1):
        handle2 = type(handle1).fromhandle(handle1.handle)
        self.assertEqual(handle1, handle2)

    def checkHandleInt(self, handle1):
        cint = handle1.toint()
        if cint == -1:
            self.skipTest(type(handle1).__name__)
        handle2 = type(handle1).fromint(cint)
        self.assertEqual(handle1, handle2)

    def checkHandleFtn(self, handle1):
        try:
            fint = handle1.py2f()
        except NotImplementedError:
            self.skipTest(type(handle1).__name__)
        if fint == -1:
            self.skipTest(type(handle1).__name__)
        handle2 = type(handle1).f2py(fint)
        self.assertEqual(handle1, handle2)

    def create(self):
        handle = None
        klass = type(self.HANDLES[0])
        if issubclass(klass, MPI.Status):
            handle = MPI.Status()
        if issubclass(klass, MPI.Datatype):
            handle = MPI.BYTE.Dup()
        if issubclass(klass, MPI.Op):
            handle = MPI.Op.Create(lambda *_: None)
        if issubclass(klass, MPI.Request):
            handle = MPI.COMM_SELF.Recv_init(bytearray(0), 0, 0)
        if issubclass(klass, MPI.Message):
            handle = MPI.MESSAGE_NULL
        if issubclass(klass, MPI.Errhandler):
            handle = MPI.ERRHANDLER_NULL
        if issubclass(klass, MPI.Info):
            handle = MPI.Info.Create()
        if issubclass(klass, MPI.Group):
            handle = MPI.COMM_SELF.Get_group()
        if issubclass(klass, MPI.Session):
            handle = MPI.Session.Init()
        if issubclass(klass, MPI.Comm):
            handle = MPI.COMM_SELF.Dup()
        if issubclass(klass, MPI.Win):
            handle = MPI.Win.Create(MPI.BOTTOM, comm=MPI.COMM_SELF)
        if issubclass(klass, MPI.File):
            import os

            name = os.devnull
            mode = MPI.MODE_RDONLY
            if os.name != "posix":
                import tempfile

                rank = MPI.COMM_WORLD.Get_rank()
                fd, name = tempfile.mkstemp(prefix=f"mpi4py-{rank}-")
                os.close(fd)
                mode |= MPI.MODE_CREATE
                mode |= MPI.MODE_DELETE_ON_CLOSE
            handle = MPI.File.Open(MPI.COMM_SELF, name, mode)
        return handle

    def destroy(self, handle):
        if handle:
            for method in ("Free", "Close", "Finalize"):
                if hasattr(handle, method):
                    getattr(handle, method)()

    def execute(self, check):
        for handle in self.HANDLES:
            check(handle)
            if not handle:
                continue
            if not hasattr(handle, "Dup"):
                continue
            handle = handle.Dup()
            try:
                check(handle)
            finally:
                handle.Free()
        try:
            handle = self.create()
        except (NotImplementedError, MPI.Exception):
            klass = type(self.HANDLES[0])
            self.skipTest(klass.__name__)
        try:
            check(handle)
        finally:
            self.destroy(handle)

    def testHandleMPI(self):
        self.execute(self.checkHandleMPI)

    def testHandleInt(self):
        self.execute(self.checkHandleInt)

    def testFortran(self):
        self.execute(self.checkHandleFtn)


class TestHandleStatus(BaseTestHandle, unittest.TestCase):
    #
    def setUp(self):
        s1 = MPI.Status()
        s2 = MPI.Status()
        s2.source = 1
        s2.tag = 2
        s2.error = MPI.ERR_OTHER
        s3 = MPI.Status()
        s3.source = 0
        s3.tag = 0
        s3.error = MPI.SUCCESS
        self.HANDLES = [s1, s2, s3]

    def testHandleMPI(self):
        pass

    def testHandleInt(self):
        pass

    @unittest.skipMPI("MPICH1")
    def testFortran(self):
        super().testFortran()

    def testFintArray(self):
        s = MPI.F_SOURCE
        t = MPI.F_TAG
        e = MPI.F_ERROR
        for status in self.HANDLES:
            try:
                f_status = status.py2f()
            except NotImplementedError:
                continue
            self.assertEqual(f_status[s], status.Get_source())
            self.assertEqual(f_status[t], status.Get_tag())
            self.assertEqual(f_status[e], status.Get_error())
            self.assertEqual(len(f_status), MPI.F_STATUS_SIZE)


class TestHandleDatatype(BaseTestHandle, unittest.TestCase):
    #
    HANDLES = [
        MPI.DATATYPE_NULL,
        MPI.CHAR,
        MPI.SHORT,
        MPI.INT,
        MPI.LONG,
        MPI.FLOAT,
        MPI.DOUBLE,
    ]


class TestHandleOp(BaseTestHandle, unittest.TestCase):
    #
    HANDLES = [
        MPI.OP_NULL,
        MPI.MAX,
        MPI.MIN,
        MPI.SUM,
        MPI.PROD,
        MPI.LAND,
        MPI.BAND,
        MPI.LOR,
        MPI.BOR,
        MPI.LXOR,
        MPI.BXOR,
        MPI.MAXLOC,
        MPI.MINLOC,
    ]


class TestHandleRequest(BaseTestHandle, unittest.TestCase):
    #
    HANDLES = [
        MPI.REQUEST_NULL,
    ]


class TestHandleMessage(BaseTestHandle, unittest.TestCase):
    #
    HANDLES = [
        MPI.MESSAGE_NULL,
        MPI.MESSAGE_NO_PROC,
    ]


class TestHandleErrhandler(BaseTestHandle, unittest.TestCase):
    #
    HANDLES = [
        MPI.ERRHANDLER_NULL,
        MPI.ERRORS_RETURN,
        MPI.ERRORS_ARE_FATAL,
        MPI.ERRORS_ABORT,
    ]


class TestHandleInfo(BaseTestHandle, unittest.TestCase):
    #
    HANDLES = [
        MPI.INFO_NULL,
    ]


class TestHandleGroup(BaseTestHandle, unittest.TestCase):
    #
    HANDLES = [
        MPI.GROUP_NULL,
        MPI.GROUP_EMPTY,
    ]


class TestHandleSession(BaseTestHandle, unittest.TestCase):
    #
    HANDLES = [
        MPI.SESSION_NULL,
    ]


class TestHandleComm(BaseTestHandle, unittest.TestCase):
    #
    HANDLES = [
        MPI.COMM_NULL,
        MPI.COMM_SELF,
        MPI.COMM_WORLD,
    ]


class TestHandleWin(BaseTestHandle, unittest.TestCase):
    #
    HANDLES = [
        MPI.WIN_NULL,
    ]


class TestHandleFile(BaseTestHandle, unittest.TestCase):
    #
    HANDLES = [
        MPI.FILE_NULL,
    ]


if __name__ == "__main__":
    unittest.main()



# -----------------------------------------------------------------------------

cdef object __UNWEIGHTED__    = <MPI_Aint>MPI_UNWEIGHTED

cdef object __WEIGHTS_EMPTY__ = <MPI_Aint>MPI_WEIGHTS_EMPTY


cdef inline bint is_UNWEIGHTED(object weights):
    return is_constant(weights, __UNWEIGHTED__)

cdef inline bint is_WEIGHTS_EMPTY(object weights):
    return is_constant(weights, __WEIGHTS_EMPTY__)


cdef inline object asarray_weights(object weights, int nweight, int **iweight):
    if weights is None:
        iweight[0] = MPI_UNWEIGHTED
        return None
    if is_UNWEIGHTED(weights):
        iweight[0] = MPI_UNWEIGHTED
        return None
    if is_WEIGHTS_EMPTY(weights):
        if nweight > 0:
            raise ValueError("empty weights but nonzero degree")
        iweight[0] = MPI_WEIGHTS_EMPTY
        return None
    return chkarray(weights, nweight, iweight)

# -----------------------------------------------------------------------------

cdef inline int comm_neighbors_count(
    MPI_Comm comm,
    int *incoming,
    int *outgoing,
) except -1:
    cdef int topo = MPI_UNDEFINED
    cdef int size=0, ndims=0, rank=0, nneighbors=0
    cdef int indegree=0, outdegree=0, weighted=0
    CHKERR( MPI_Topo_test(comm, &topo) )
    if topo == MPI_UNDEFINED:                 # ~> unreachable
        CHKERR( MPI_Comm_size(comm, &size) )  # ~> unreachable
        indegree = outdegree = size           # ~> unreachable
    elif topo == MPI_CART:
        CHKERR( MPI_Cartdim_get(comm, &ndims) )
        indegree = outdegree = <int>2*ndims
    elif topo == MPI_GRAPH:
        CHKERR( MPI_Comm_rank(comm, &rank) )
        CHKERR( MPI_Graph_neighbors_count(
                comm, rank, &nneighbors) )
        indegree = outdegree = nneighbors
    elif topo == MPI_DIST_GRAPH:
        CHKERR( MPI_Dist_graph_neighbors_count(
                comm, &indegree, &outdegree, &weighted) )
    if incoming != NULL: incoming[0] = indegree
    if outgoing != NULL: outgoing[0] = outdegree
    return 0

# -----------------------------------------------------------------------------

cdef int    commlock_keyval   = MPI_KEYVAL_INVALID
cdef object commlock_lock     = Lock()
cdef dict   commlock_registry = {}


cdef inline int commlock_free_cb(
    MPI_Comm comm,
) except MPI_ERR_UNKNOWN with gil:
    cdef object key = <Py_uintptr_t>comm
    with commlock_lock:
        if key in commlock_registry:
            del commlock_registry[key]
    return MPI_SUCCESS


@cython.linetrace(False)
@cython.callspec("MPIAPI")
cdef int commlock_free_fn(
    MPI_Comm comm,
    int keyval,
    void *attrval,
    void *xstate,
) noexcept nogil:
    <void> keyval   # unused
    <void> attrval  # unused
    <void> xstate   # unused
    if comm == MPI_COMM_SELF:  <void>MPI_Comm_free_keyval(&commlock_keyval)
    if not Py_IsInitialized(): return MPI_SUCCESS
    if not py_module_alive():  return MPI_SUCCESS
    return commlock_free_cb(comm)


cdef inline dict commlock_table(MPI_Comm comm):
    cdef int found = 0
    cdef void *attrval = NULL
    cdef dict table
    if commlock_keyval == MPI_KEYVAL_INVALID:
        CHKERR( MPI_Comm_create_keyval(
            MPI_COMM_NULL_COPY_FN,
            commlock_free_fn,
            &commlock_keyval, NULL) )
        table = {}
        CHKERR( MPI_Comm_set_attr(
            MPI_COMM_SELF, commlock_keyval, <void*> table) )
        commlock_registry[<Py_uintptr_t>MPI_COMM_SELF] = table
    CHKERR( MPI_Comm_get_attr(
        comm, commlock_keyval, &attrval, &found) )
    if not found:
        table = {}
        CHKERR( MPI_Comm_set_attr(
            comm, commlock_keyval, <void*> table) )
        commlock_registry[<Py_uintptr_t>comm] = table
    elif PYPY:
        table = commlock_registry[<Py_uintptr_t>comm]  # ~> pypy
    else:
        table = <dict> attrval
    return table


cdef inline object PyMPI_Lock(MPI_Comm comm, object key):
    cdef dict table
    cdef object lock
    with commlock_lock:
        table = commlock_table(comm)
        try:
            lock = table[key]
        except KeyError:
            lock = Lock()
            table[key] = lock
        return lock


cdef inline object PyMPI_Lock_table(MPI_Comm comm):
    with commlock_lock:
        return commlock_table(comm)


def _comm_lock(Comm comm: Comm, object key: Hashable | None = None) -> Lock:
    """
    Create/get communicator lock.
    """
    return PyMPI_Lock(comm.ob_mpi, key)


def _comm_lock_table(Comm comm: Comm) -> dict[Hashable, Lock]:
    """
    Internal communicator lock table.
    """
    return PyMPI_Lock_table(comm.ob_mpi)

# -----------------------------------------------------------------------------



# -----------------------------------------------------------------------------

cdef extern from * nogil:  # "stddef.h"
    ctypedef unsigned int wchar_t

cdef extern from * nogil:  # "stdlib.h"
    const char *getenv(const char *)

cdef extern from * nogil:  # "string.h"
    int strcmp(const char *, const char *)
    int strncmp(const char *, const char *, size_t)

cdef extern from * nogil:  # "string.h"
    int memcmp(const void *, const void *, size_t)
    void *memset(void *, int, size_t)
    void *memcpy(void *, const void *, size_t)
    void *memmove(void *, const void *, size_t)

cdef extern from * nogil:  # "stdio.h"
    ctypedef struct FILE
    FILE *stdin, *stdout, *stderr
    int fprintf(FILE *, char *, ...)
    int fflush(FILE *)

# -----------------------------------------------------------------------------



import argparse
import collections
import copy
import fnmatch
import json

parser = argparse.ArgumentParser()
parser.add_argument("--os", nargs="*")
parser.add_argument("--py", nargs="*")
opts = parser.parse_args()


def py(py, x, y_min, y_max, abi=""):
    return [f"{py}{x}{y}{abi}" for y in range(y_min, y_max + 1)]


def cp3(y_min=8, y_max=14):
    return py("cp", 3, y_min, y_max)


def cp3t(y_min=13, y_max=14):
    return py("cp", 3, y_min, y_max, abi="t")


def abi3(y_min=10, y_max=10):
    return py("cp", 3, y_min, y_max, abi="-abi3")


def pp3(y_min=11, y_max=11):
    return py("pp", 3, y_min, y_max)


OS_ARCH_PY = {
    "Linux": {
        "aarch64": cp3() + cp3t() + abi3() + pp3(),
        "x86_64": cp3() + cp3t() + abi3() + pp3(),
    },
    "macOS": {
        "arm64": cp3() + cp3t() + abi3() + pp3(),
        "x86_64": cp3() + cp3t() + abi3() + pp3(),
    },
    "Windows": {
        "AMD64": cp3() + cp3t() + abi3() + pp3(),
    },
}

MPI_ABI_POSIX = [
    "mpich",
    "openmpi",
]
MPI_ABI_WINNT = [
    "impi",
    "msmpi",
]
MPI_ABI = {
    "Linux": MPI_ABI_POSIX.copy(),
    "macOS": MPI_ABI_POSIX.copy(),
    "Windows": MPI_ABI_WINNT.copy(),
}

GHA_RUNNER = {
    "Linux": {
        "aarch64": "ubuntu-24.04-arm",
        "x86_64": "ubuntu-24.04",
        None: "ubuntu-latest",
    },
    "macOS": {
        "arm64": "macos-15",
        "x86_64": "macos-15-intel",
        None: "macos-latest",
    },
    "Windows": {
        "AMD64": "windows-2022",
        None: "windows-latest",
    },
}

os_arch_py = copy.deepcopy(OS_ARCH_PY)
if opts.os and not set(opts.os) & {"*", "all"}:
    select = collections.defaultdict(list)
    for entry in opts.os:
        for sep in "=+/@":
            entry = entry.replace(sep, "-")
        os, _, arch = entry.partition("-")
        assert os in OS_ARCH_PY  # noqa: S101
        if arch and arch not in ("*", "all"):
            assert arch in OS_ARCH_PY[os]  # noqa: S101
            if arch not in select[os]:
                select[os].append(arch)
        else:
            for arch in OS_ARCH_PY[os]:
                if arch not in select[os]:
                    select[os].append(arch)
    os_arch_py = collections.defaultdict(dict)
    for os in select:
        for arch in select[os]:
            os_arch_py[os][arch] = OS_ARCH_PY[os][arch].copy()
if opts.py and not set(opts.py) & {"*", "all"}:
    for os in os_arch_py:
        for arch in os_arch_py[os]:
            xplist = os_arch_py[os][arch] + pp3()
            select = []
            for pat in opts.py:
                for xp in fnmatch.filter(xplist, pat):
                    if xp not in select:
                        select.append(xp)
            if "abi3" in opts.py:
                for xp in abi3():
                    if xp not in select:
                        select.append(xp)
            os_arch_py[os][arch][:] = select

matrix_build = [
    {
        "os": os,
        "arch": arch,
        "py": py.partition("-")[0],
        "py-sabi": py.partition("-")[2],
        "mpi-abi": mpi_abi,
        "runner": GHA_RUNNER[os][arch],
    }
    for os in os_arch_py
    for arch in os_arch_py[os]
    for py in os_arch_py[os][arch]
    for mpi_abi in MPI_ABI[os]
]

matrix_merge = [
    {
        "os": os,
        "arch": arch,
        "runner": GHA_RUNNER[os][None],
    }
    for os in os_arch_py
    for arch in os_arch_py[os]
]

matrix_test = []
for build in matrix_build:
    os = build["os"]
    arch = build["arch"]
    pytag = build["py"]
    mpi_abi = build["mpi-abi"]
    py_sabi = build["py-sabi"]
    runner = GHA_RUNNER[os][arch]
    mpilist = [mpi_abi]
    if (os, arch, mpi_abi) == ("Linux", "x86_64", "mpich"):
        mpilist.insert(0, "impi")
    if py_sabi:
        pyvers = [py for py in cp3() if int(py[3:]) >= int(pytag[3:])]
        pylist = [py[2:3] + "." + py[3:] for py in pyvers]
    else:
        pyimpl = "pypy" if pytag.startswith("pp") else ""
        pylist = [pyimpl + pytag[2:3] + "." + pytag[3:]]
    matrix_test += [
        {
            "mpi": mpi,
            "py": py,
            "py-sabi": py_sabi,
            "os": os,
            "arch": arch,
            "runner": runner,
        }
        for py in pylist
        for mpi in mpilist
    ]

print(f"matrix-build={json.dumps(matrix_build)}")
print(f"matrix-merge={json.dumps(matrix_merge)}")
print(f"matrix-test={json.dumps(matrix_test)}")



import contextlib
import sys

import arrayimpl
import mpiunittest as unittest

from mpi4py import MPI

scalar = arrayimpl.scalar

typemap = MPI.Datatype.fromcode


def mkzeros(n):
    return bytearray(n)


def memzero(m):
    try:
        m[:] = 0
    except IndexError:  # cffi buffer
        m[0 : len(m)] = b"\0" * len(m)


@contextlib.contextmanager
def win_lock(win, rank, *args, **kwargs):
    win.Lock(rank, *args, **kwargs)
    try:
        yield
    finally:
        win.Unlock(rank)


@contextlib.contextmanager
def win_lock_all(win, *args, **kwargs):
    win.Lock_all(*args, **kwargs)
    try:
        yield
    finally:
        win.Unlock_all()


class BaseTestRMA:
    #
    COMM = MPI.COMM_NULL
    INFO = MPI.INFO_NULL

    def setUp(self):
        nbytes = 100 * MPI.DOUBLE.size
        try:
            self.mpi_memory = MPI.Alloc_mem(nbytes)
            self.memory = self.mpi_memory
            memzero(self.memory)
        except MPI.Exception:
            self.mpi_memory = None
            self.memory = bytearray(nbytes)
        self.WIN = MPI.Win.Create(self.memory, 1, self.INFO, self.COMM)

    def tearDown(self):
        self.WIN.Free()
        if self.mpi_memory:
            MPI.Free_mem(self.mpi_memory)

    def testPutGet(self):
        group = self.WIN.Get_group()
        size = group.Get_size()
        group.Free()
        for array, typecode in arrayimpl.loop():
            with arrayimpl.test(self):
                if unittest.is_mpi_gpu("mvapich", array):
                    continue
                for count in range(10):
                    for rank in range(size):
                        with self.subTest(rank=rank, count=count):
                            sbuf = array(range(count), typecode)
                            rbuf = array(-1, typecode, count + 1)
                            #
                            self.WIN.Fence()
                            self.WIN.Put(sbuf.as_mpi(), rank)
                            self.WIN.Fence()
                            self.WIN.Get(rbuf.as_mpi_c(count), rank)
                            self.WIN.Fence()
                            for i in range(count):
                                self.assertEqual(sbuf[i], scalar(i))
                                self.assertEqual(rbuf[i], scalar(i))
                            self.assertEqual(rbuf[-1], scalar(-1))
                            #
                            sbuf = array(range(count), typecode)
                            rbuf = array(-1, typecode, count + 1)
                            target = sbuf.itemsize
                            self.WIN.Fence()
                            self.WIN.Put(sbuf.as_mpi(), rank, target)
                            self.WIN.Fence()
                            self.WIN.Get(rbuf.as_mpi_c(count), rank, target)
                            self.WIN.Fence()
                            for i in range(count):
                                self.assertEqual(sbuf[i], scalar(i))
                                self.assertEqual(rbuf[i], scalar(i))
                            self.assertEqual(rbuf[-1], scalar(-1))
                            #
                            sbuf = array(range(count), typecode)
                            rbuf = array(-1, typecode, count + 1)
                            datatype = typemap(typecode)
                            target = (sbuf.itemsize, count, datatype)
                            self.WIN.Fence()
                            self.WIN.Put(sbuf.as_mpi(), rank, target)
                            self.WIN.Fence()
                            self.WIN.Get(rbuf.as_mpi_c(count), rank, target)
                            self.WIN.Fence()
                            for i in range(count):
                                self.assertEqual(sbuf[i], scalar(i))
                                self.assertEqual(rbuf[i], scalar(i))
                            self.assertEqual(rbuf[-1], scalar(-1))

    def testAccumulate(self):
        group = self.WIN.Get_group()
        size = group.Get_size()
        group.Free()
        for array, typecode in arrayimpl.loop():
            with arrayimpl.test(self):
                if unittest.is_mpi_gpu("openmpi", array):
                    continue
                if unittest.is_mpi_gpu("mvapich", array):
                    continue
                if typecode == "?":
                    continue
                if typecode in "FDG":
                    continue
                for count in range(10):
                    for rank in range(size):
                        with self.subTest(rank=rank, count=count):
                            sbuf = array(range(count), typecode)
                            rbuf = array(-1, typecode, count + 1)
                            for op in (MPI.SUM, MPI.PROD, MPI.MAX, MPI.MIN):
                                self.WIN.Fence()
                                self.WIN.Accumulate(sbuf.as_mpi(), rank, op=op)
                                self.WIN.Fence()
                                self.WIN.Get(rbuf.as_mpi_c(count), rank)
                                self.WIN.Fence()
                                for i in range(count):
                                    self.assertEqual(sbuf[i], scalar(i))
                                    self.assertNotEqual(rbuf[i], scalar(-1))
                                self.assertEqual(rbuf[-1], scalar(-1))

    @unittest.skipMPI("openmpi(>=1.10,<1.11)")
    def testGetAccumulate(self):
        group = self.WIN.Get_group()
        size = group.Get_size()
        rank = group.Get_rank()
        group.Free()
        self.WIN.Fence()
        obuf = MPI.Alloc_mem(1)
        memzero(obuf)
        rbuf = MPI.Alloc_mem(1)
        memzero(rbuf)
        try:
            try:
                self.WIN.Get_accumulate(
                    [obuf, 0, MPI.BYTE], [rbuf, 0, MPI.BYTE], rank
                )
                self.WIN.Fence()
            finally:
                MPI.Free_mem(obuf)
                MPI.Free_mem(rbuf)
        except NotImplementedError:
            self.skipTest("mpi-win-get_accumulate")
        self.WIN.Fence()
        for array, typecode in arrayimpl.loop():
            with arrayimpl.test(self):
                if unittest.is_mpi_gpu("openmpi", array):
                    continue
                if unittest.is_mpi_gpu("mvapich", array):
                    continue
                if typecode == "?":
                    continue
                if typecode in "FDG":
                    continue
                for count in range(10):
                    for rank in range(size):
                        with self.subTest(rank=rank, count=count):
                            ones = array([1] * count, typecode)
                            sbuf = array(range(count), typecode)
                            rbuf = array(-1, typecode, count + 1)
                            gbuf = array(-1, typecode, count + 1)
                            for op in (
                                MPI.SUM,
                                MPI.PROD,
                                MPI.MAX,
                                MPI.MIN,
                                MPI.REPLACE,
                                MPI.NO_OP,
                            ):
                                with win_lock(self.WIN, rank):
                                    self.WIN.Put(ones.as_mpi(), rank)
                                    self.WIN.Flush(rank)
                                    self.WIN.Get_accumulate(
                                        sbuf.as_mpi(),
                                        rbuf.as_mpi_c(count),
                                        rank,
                                        op=op,
                                    )
                                    self.WIN.Flush(rank)
                                    self.WIN.Get(gbuf.as_mpi_c(count), rank)
                                    self.WIN.Flush(rank)
                                #
                                for i in range(count):
                                    self.assertEqual(sbuf[i], scalar(i))
                                    self.assertEqual(rbuf[i], scalar(1))
                                    self.assertEqual(gbuf[i], scalar(op(1, i)))
                                self.assertEqual(rbuf[-1], scalar(-1))
                                self.assertEqual(gbuf[-1], scalar(-1))

    def testFetchAndOp(self):
        group = self.WIN.Get_group()
        size = group.Get_size()
        rank = group.Get_rank()
        group.Free()
        self.WIN.Fence()
        blen = MPI.INT.Get_size()
        obuf = MPI.Alloc_mem(blen)
        memzero(obuf)
        rbuf = MPI.Alloc_mem(blen)
        memzero(rbuf)
        try:
            try:
                self.WIN.Fetch_and_op(
                    [obuf, 1, MPI.INT], [rbuf, 1, MPI.INT], rank
                )
                self.WIN.Fence()
            finally:
                MPI.Free_mem(obuf)
                MPI.Free_mem(rbuf)
        except NotImplementedError:
            self.skipTest("mpi-win-fetch_and_op")
        self.WIN.Fence()
        for array, typecode in arrayimpl.loop():
            with arrayimpl.test(self):
                if unittest.is_mpi_gpu("openmpi", array):
                    continue
                if unittest.is_mpi_gpu("mvapich", array):
                    continue
                if typecode == "?":
                    continue
                if typecode in "FDG":
                    continue
                obuf = array(+1, typecode)
                rbuf = array(-1, typecode, 2)
                datatype = typemap(typecode)
                for op in (
                    MPI.SUM,
                    MPI.PROD,
                    MPI.MAX,
                    MPI.MIN,
                    MPI.REPLACE,
                    MPI.NO_OP,
                ):
                    for rank in range(size):
                        for disp in range(3):
                            with self.subTest(disp=disp, rank=rank):
                                with win_lock(self.WIN, rank):
                                    self.WIN.Fetch_and_op(
                                        obuf.as_mpi(),
                                        rbuf.as_mpi_c(1),
                                        rank,
                                        disp * datatype.size,
                                        op=op,
                                    )
                                self.assertEqual(rbuf[1], scalar(-1))
        big = bytearray(MPI.INT.Get_size())
        buf1 = bytearray(1)
        buf2 = bytearray(1)
        with win_lock(self.WIN, rank):
            self.WIN.Fetch_and_op(buf1, buf2, rank, op=MPI.NO_OP)
        with self.assertRaises(ValueError):
            with win_lock(self.WIN, rank):
                self.WIN.Fetch_and_op(big, buf2, rank, op=MPI.NO_OP)
        with self.assertRaises(ValueError):
            with win_lock(self.WIN, rank):
                self.WIN.Fetch_and_op(buf1, big, rank, op=MPI.NO_OP)
        with self.assertRaises(ValueError):
            with win_lock(self.WIN, rank):
                self.WIN.Fetch_and_op(buf1, [big, MPI.INT], rank, op=MPI.NO_OP)

    @unittest.skipMPI("mpich(>=4.0,<4.1)", sys.platform == "darwin")
    def testCompareAndSwap(self):
        group = self.WIN.Get_group()
        size = group.Get_size()
        rank = group.Get_rank()
        group.Free()
        self.WIN.Fence()
        obuf = MPI.Alloc_mem(1)
        memzero(obuf)
        cbuf = MPI.Alloc_mem(1)
        memzero(cbuf)
        rbuf = MPI.Alloc_mem(1)
        memzero(rbuf)
        try:
            try:
                self.WIN.Compare_and_swap(
                    [obuf, 1, MPI.BYTE],
                    [cbuf, 1, MPI.BYTE],
                    [rbuf, 1, MPI.BYTE],
                    rank,
                    0,
                )
                self.WIN.Fence()
            finally:
                MPI.Free_mem(obuf)
                MPI.Free_mem(cbuf)
                MPI.Free_mem(rbuf)
        except NotImplementedError:
            self.skipTest("mpi-win-compare_and_swap")
        self.WIN.Fence()
        for array, typecode in arrayimpl.loop():
            with arrayimpl.test(self):
                if unittest.is_mpi_gpu("openmpi", array):
                    continue
                if unittest.is_mpi_gpu("mvapich", array):
                    continue
                if typecode in "fdg":
                    continue
                if typecode in "FDG":
                    continue
                obuf = array(+1, typecode)
                cbuf = array(0, typecode)
                rbuf = array(-1, typecode, 2)
                datatype = typemap(typecode)
                for rank in range(size):
                    for disp in range(3):
                        with self.subTest(disp=disp, rank=rank):
                            with win_lock(self.WIN, rank):
                                self.WIN.Compare_and_swap(
                                    obuf.as_mpi(),
                                    cbuf.as_mpi(),
                                    rbuf.as_mpi_c(1),
                                    rank,
                                    disp * datatype.size,
                                )
                            self.assertEqual(rbuf[1], scalar(-1))
        big = bytearray(MPI.INT.Get_size())
        buf1 = bytearray(1)
        buf2 = bytearray(1)
        buf3 = bytearray(1)
        with win_lock(self.WIN, rank):
            self.WIN.Compare_and_swap(buf1, buf2, buf3, rank)
        with self.assertRaises(ValueError):
            with win_lock(self.WIN, rank):
                self.WIN.Compare_and_swap(big, buf2, buf3, rank)
        with self.assertRaises(ValueError):
            with win_lock(self.WIN, rank):
                self.WIN.Compare_and_swap(buf1, big, buf3, rank)
        with self.assertRaises(ValueError):
            with win_lock(self.WIN, rank):
                self.WIN.Compare_and_swap(buf1, buf2, big, rank)
        with self.assertRaises(ValueError):
            with win_lock(self.WIN, rank):
                self.WIN.Compare_and_swap(buf1, [big, MPI.INT], buf3, rank)
        with self.assertRaises(ValueError):
            with win_lock(self.WIN, rank):
                self.WIN.Compare_and_swap(buf1, buf2, [big, MPI.INT], rank)

    def testPutProcNull(self):
        self.WIN.Fence()
        self.WIN.Put(None, MPI.PROC_NULL, None)
        self.WIN.Fence()

    def testGetProcNull(self):
        self.WIN.Fence()
        self.WIN.Get(None, MPI.PROC_NULL, None)
        self.WIN.Fence()

    def testAccumulateProcNullReplace(self):
        self.WIN.Fence()
        zeros = mkzeros(8)
        self.WIN.Fence()
        self.WIN.Accumulate([zeros, MPI.INT], MPI.PROC_NULL, None, MPI.REPLACE)
        self.WIN.Fence()
        self.WIN.Accumulate([zeros, MPI.INT], MPI.PROC_NULL, None, MPI.REPLACE)
        self.WIN.Fence()

    def testAccumulateProcNullSum(self):
        self.WIN.Fence()
        zeros = mkzeros(8)
        self.WIN.Fence()
        self.WIN.Accumulate([zeros, MPI.INT], MPI.PROC_NULL, None, MPI.SUM)
        self.WIN.Fence()
        self.WIN.Accumulate([None, MPI.INT], MPI.PROC_NULL, None, MPI.SUM)
        self.WIN.Fence()

    def testGetAccumulateProcNull(self):
        obuf = [mkzeros(8), 0, MPI.INT]
        rbuf = [mkzeros(8), 0, MPI.INT]
        self.WIN.Fence()
        try:
            self.WIN.Get_accumulate(obuf, rbuf, MPI.PROC_NULL)
        except NotImplementedError:
            self.skipTest("mpi-win-get_accumulate")
        self.WIN.Fence()

    # def testFetchAndOpProcNull(self):
    #     obuf = cbuf = rbuf = None
    #     self.WIN.Fence()
    #     try:
    #         self.WIN.Fetch_and_op(obuf, rbuf, MPI.PROC_NULL, 0)
    #     except NotImplementedError:
    #         self.skipTest('mpi-win-fetch_and_op')
    #     self.WIN.Fence()

    # def testCompareAndSwapProcNull(self):
    #     obuf = cbuf = rbuf = None
    #     self.WIN.Fence()
    #     try:
    #         self.WIN.Compare_and_swap(obuf, cbuf, rbuf, MPI.PROC_NULL, 0)
    #     except NotImplementedError:
    #         self.skipTest('mpi-win-compare_and_swap')
    #     self.WIN.Fence()

    def testFence(self):
        win = self.WIN
        LMODE = [
            0,
            MPI.MODE_NOSTORE,
            MPI.MODE_NOPUT,
            MPI.MODE_NOSTORE | MPI.MODE_NOPUT,
        ]
        GMODE = [
            0,
            MPI.MODE_NOPRECEDE,
            MPI.MODE_NOSUCCEED,
        ]
        win.Fence()
        for lmode in LMODE:
            for gmode in GMODE:
                assertion = lmode | gmode
                win.Fence(assertion)
        win.Fence()

    @unittest.skipMPI("openmpi(==1.8.1)")
    def testFenceAll(self):
        win = self.WIN
        assertion = 0
        modes = [
            0,
            MPI.MODE_NOSTORE,
            MPI.MODE_NOPUT,
            MPI.MODE_NOPRECEDE,
            MPI.MODE_NOSUCCEED,
        ]
        win.Fence()
        for mode in modes:
            win.Fence(mode)
            assertion |= mode
            win.Fence(assertion)
        win.Fence()

    @unittest.skipMPI("openmpi(==1.8.6)")
    def testStartComplete(self):
        self.WIN.Start(MPI.GROUP_EMPTY)
        self.WIN.Complete()

    @unittest.skipMPI("openmpi(==1.8.6)")
    def testPostWait(self):
        self.WIN.Post(MPI.GROUP_EMPTY)
        self.WIN.Wait()

    @unittest.skipMPI("openmpi(==1.8.7)")
    @unittest.skipMPI("openmpi(==1.8.6)")
    def testStartCompletePostWait(self):
        win = self.WIN
        wingroup = win.Get_group()
        size = wingroup.Get_size()
        rank = wingroup.Get_rank()
        if size < 2:
            return wingroup.Free()
        if rank == 0:
            group = wingroup.Excl([0])
            win.Start(group)
            win.Complete()
            win.Post(group)
            win.Wait()
            group.Free()
        else:
            group = wingroup.Incl([0])
            win.Post(group)
            win.Wait()
            win.Start(group)
            win.Complete()
            group.Free()
        wingroup.Free()
        return None

    @unittest.skipMPI("openmpi(==1.8.7)")
    @unittest.skipMPI("openmpi(==1.8.6)")
    def testStartCompletePostTest(self):
        comm = self.COMM
        win = self.WIN
        wingroup = win.Get_group()
        size = wingroup.Get_size()
        rank = wingroup.Get_rank()
        if size < 2:
            return wingroup.Free()
        if rank == 0:
            group = wingroup.Excl([0])
            win.Start(group)
            comm.Barrier()
            win.Complete()
            comm.Barrier()
            group.Free()
        else:
            group = wingroup.Incl([0])
            win.Post(group)
            flag = win.Test()
            self.assertFalse(flag)
            comm.Barrier()
            comm.Barrier()
            flag = win.Test()
            self.assertTrue(flag)
            group.Free()
        wingroup.Free()
        return None

    @unittest.skipMPI("MPI(<3.0)")
    def testSync(self):
        win = self.WIN
        comm = self.COMM
        rank = comm.Get_rank()
        with win_lock(win, rank):
            win.Sync()
        comm.Barrier()

    @unittest.skipMPI("MPI(<3.0)")
    def testFlush(self):
        win = self.WIN
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        #
        for i in range(size):
            with win_lock(win, i):
                win.Flush(i)
        comm.Barrier()
        for i in range(size):
            if i == rank:
                with win_lock_all(win):
                    win.Flush_all()
            comm.Barrier()
        #
        for i in range(size):
            with win_lock(win, i):
                win.Flush_local(i)
        comm.Barrier()
        for i in range(size):
            if i == rank:
                with win_lock_all(win):
                    win.Flush_local_all()
            comm.Barrier()


class TestRMASelf(BaseTestRMA, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


class TestRMAWorld(BaseTestRMA, unittest.TestCase):
    #
    COMM = MPI.COMM_WORLD


try:
    MPI.Win.Create(None, 1, MPI.INFO_NULL, MPI.COMM_SELF).Free()
except (NotImplementedError, MPI.Exception):
    unittest.disable(BaseTestRMA, "mpi-rma")


if __name__ == "__main__":
    unittest.main()



import mpiunittest as unittest

from mpi4py import MPI


class TestSession(unittest.TestCase):
    #
    def testSessionInit(self):
        session = MPI.Session()
        self.assertFalse(session)
        self.assertEqual(session, MPI.SESSION_NULL)
        session = MPI.Session.Init()
        self.assertTrue(session)
        self.assertNotEqual(session, MPI.SESSION_NULL)
        self.assertEqual(session, MPI.Session(session))
        session.Finalize()

    def testSessionGetInfo(self):
        session = MPI.Session.Init()
        info = session.Get_info()
        info.Free()
        session.Finalize()

    def testSessionPsets(self):
        session = MPI.Session.Init()
        num_psets = session.Get_num_psets()
        for n in range(num_psets):
            name = session.Get_nth_pset(n)
            self.assertGreater(len(name), 0)
        session.Finalize()

    def testSessionPsetInfo(self):
        session = MPI.Session.Init()
        num_psets = session.Get_num_psets()
        for n in range(num_psets):
            name = session.Get_nth_pset(n)
            info = session.Get_pset_info(name)
            info.Free()
        session.Finalize()

    def testSessionPsetGroup(self):
        session = MPI.Session.Init()
        num_psets = session.Get_num_psets()
        for n in range(num_psets):
            name = session.Get_nth_pset(n)
            try:
                group = session.Create_group(name)
                group.Free()
                group = MPI.Group.Create_from_session_pset(session, name)
                group.Free()
            except MPI.Exception as exc:  # openmpi
                UNSUPPORTED = MPI.ERR_OTHER
                if exc.Get_error_class() != UNSUPPORTED:
                    raise
        session.Finalize()

    def testSessionSELF(self):
        session = MPI.Session.Init()
        name = "mpi://SELF"
        info = session.Get_pset_info(name)
        self.assertEqual(info.Get("mpi_size"), "1")
        info.Free()
        group = session.Create_group(name)
        self.assertEqual(group.Get_rank(), 0)
        self.assertEqual(group.Get_size(), 1)
        group.Free()
        session.Finalize()

    def testSessionWORLD(self):
        comm = MPI.COMM_WORLD
        session = MPI.Session.Init()
        name = "mpi://WORLD"
        info = session.Get_pset_info(name)
        size = comm.Get_size()
        self.assertEqual(info.Get("mpi_size"), str(size))
        info.Free()
        group = session.Create_group(name)
        self.assertEqual(group.Get_size(), comm.Get_size())
        self.assertEqual(group.Get_rank(), comm.Get_rank())
        group.Free()
        session.Finalize()

    def testBuffering(self):
        session = MPI.Session.Init()
        buf = MPI.Alloc_mem((1 << 16) + MPI.BSEND_OVERHEAD)
        try:
            with self.catchNotImplementedError(4, 1):
                session.Attach_buffer(buf)
            with self.catchNotImplementedError(4, 1):
                session.Flush_buffer()
            with self.catchNotImplementedError(4, 1):
                session.Iflush_buffer().Wait()
        finally:
            with self.catchNotImplementedError(4, 1):
                oldbuf = session.Detach_buffer()
                self.assertEqual(oldbuf.address, buf.address)
                self.assertEqual(oldbuf.nbytes, buf.nbytes)
            MPI.Free_mem(buf)
            with self.catchNotImplementedError(4, 1):
                session.Attach_buffer(MPI.BUFFER_AUTOMATIC)
                bufauto = session.Detach_buffer()
                self.assertEqual(bufauto, MPI.BUFFER_AUTOMATIC)
            session.Finalize()

    def testPickle(self):
        from pickle import dumps, loads

        session = MPI.Session.Init()
        with self.assertRaises(ValueError):
            loads(dumps(session))
        session.Finalize()


try:
    MPI.Session.Init().Finalize()
except NotImplementedError:
    unittest.disable(TestSession, "mpi-session")


if __name__ == "__main__":
    unittest.main()



"""https://github.com/scientific-python/spin"""  # noqa: D400

import os
import pathlib
import shlex
import shutil
import sys

import click
from spin.cmds.util import run

PYTHON = shlex.split(os.environ.get("PYTHON", sys.executable))
MPIEXEC = shlex.split(os.environ.get("MPIEXEC", "mpiexec"))
SHELL = shlex.split(os.environ.get("SHELL", "sh"))


@click.command()
@click.option("-c", "--clean", is_flag=True, help="Clean build directory.")
@click.option("-f", "--force", is_flag=True, help="Force build everything.")
@click.option("-q", "--quiet", is_flag=True, help="Run quietly.")
def build(clean, force, quiet):
    """🔧 Build in-place."""
    opt_force = ["--force"] if force else []
    opt_quiet = ["--quiet"] if quiet else []
    if clean:
        run([*PYTHON, "setup.py", *opt_quiet, "clean", "--all"])
    run([*PYTHON, "setup.py", *opt_quiet, "build", *opt_force, "--inplace"])


@click.command()
@click.option("-n", default=2, help="Number of MPI processes")
@click.option("-s", "--singleton", is_flag=True, help="Singleton mode.")
@click.pass_context
def check(ctx, n, singleton):
    """🔧 Check in-place."""
    ctx.invoke(build, quiet=True)
    launcher = [] if singleton else [*MPIEXEC, "-n", f"{n}"]
    run([*PYTHON, "-m", "mpi4py", "--prefix"])
    run([*PYTHON, "-m", "mpi4py", "--mpi-library"])
    run([*launcher, *PYTHON, "-m", "mpi4py.bench", "ringtest"])


@click.command()
@click.option("-n", default=1, help="Number of MPI processes")
@click.option("-s", "--singleton", is_flag=True, help="Singleton mode.")
@click.argument("test_args", nargs=-1)
@click.pass_context
def test(ctx, n, singleton, test_args):
    """🔧 Test in-place."""
    ctx.invoke(build, quiet=True)
    launcher = [] if singleton else [*MPIEXEC, "-n", f"{n}"]
    run([*launcher, *PYTHON, "test/main.py", "--inplace", *test_args])


def _get_topdir():
    script = pathlib.Path(__file__).resolve()
    return script.parent.parent


def _get_site_packages():
    return _get_topdir() / "src"


def _set_pythonpath(path, quiet=False):
    pythonpath = os.environ.get("PYTHONPATH")
    if pythonpath is not None:
        pythonpath = f"{path}{os.pathsep}{pythonpath}"
    else:
        pythonpath = f"{path}"
    os.environ["PYTHONPATH"] = pythonpath
    if not quiet:
        click.secho(
            f'$ export PYTHONPATH="{pythonpath}"',
            bold=True,
            fg="bright_blue",
        )
    return path


def _setup_environment(ctx, quiet=False):
    path = _get_site_packages()
    ctx.invoke(build, quiet=True)
    _set_pythonpath(path, quiet=quiet)


def _run_check_commands(*commands):
    for cmd, *args in commands:
        if shutil.which(cmd) is not None:
            run([cmd, *args])
        else:
            click.secho(
                f"{cmd}: command not found...",
                bold=True,
                fg="bright_red",
            )


@click.command()
@click.pass_context
def lint(ctx):
    """🔦 Lint-check sources with linters."""
    _setup_environment(ctx)
    _run_check_commands(
        ["ruff", "check", "--quiet"],
        ["ruff", "format", "--quiet", "--diff"],
        ["flake8"],
        ["pylint", "mpi4py"],
        ["cython-lint", "."],
        ["codespell"],
        ["yamllint", "."],
    )


@click.command()
@click.pass_context
def type(ctx):  # noqa: A001
    """🦆 Type-check sources with mypy."""
    _setup_environment(ctx)
    _run_check_commands(
        ["stubtest", "--ignore-disjoint-bases", "mpi4py"],
        ["mypy", "-p", "mpi4py"],
    )


@click.command(context_settings={"ignore_unknown_options": True})
@click.argument("shell_args", nargs=-1)
@click.pass_context
def shell(ctx, shell_args):
    """💻 Launch shell with PYTHONPATH set."""
    _setup_environment(ctx)
    run([*SHELL, *shell_args], replace=True)


@click.command(context_settings={"ignore_unknown_options": True})
@click.argument("ipython_args", nargs=-1)
@click.pass_context
def ipython(ctx, ipython_args):
    """🐍 Launch ipython with PYTHONPATH set."""
    _setup_environment(ctx)
    ipython = [*PYTHON, "-m", "IPython"]
    run([*ipython, *ipython_args], replace=True)


@click.command(context_settings={"ignore_unknown_options": True})
@click.option("-n", default=1, help="Number of MPI processes")
@click.argument("mpiexec_args", nargs=-1)
@click.pass_context
def mpiexec(ctx, n, mpiexec_args):
    """🏁 Run mpiexec with PYTHONPATH set."""
    _setup_environment(ctx)
    run([*MPIEXEC, "-n", f"{n}", *mpiexec_args], replace=True)


@click.command(context_settings={"ignore_unknown_options": True})
@click.option("-n", default=1, help="Number of MPI processes")
@click.argument("mpi4py_args", nargs=-1)
@click.pass_context
def mpi4py(ctx, n, mpi4py_args):
    """🏁 Run mpi4py with PYTHONPATH set."""
    _setup_environment(ctx)
    mpi4py = [*MPIEXEC, "-n", f"{n}", *PYTHON, "-m", "mpi4py"]
    mpi4py_args = mpi4py_args or ["--help"]
    run([*mpi4py, *mpi4py_args], replace=True)


@click.command()
@click.option("-e", "--editable", is_flag=True, help="Editable mode.")
@click.option("-q", "--quiet", is_flag=True, help="Run quietly.")
@click.argument("pip_args", nargs=-1)
def install(editable, quiet, pip_args):
    """🔧 Install package."""
    pip = [*PYTHON, "-m", "pip"]
    pip_args = [*pip_args]
    if quiet:
        pip_args.append("--quiet")
    if editable:
        pip_args.append("--editable")
    run([*pip, "install", *pip_args, "."])


@click.command()
@click.option("-q", "--quiet", is_flag=True, help="Run quietly.")
@click.argument("pip_args", nargs=-1)
@click.pass_context
def editable(ctx, quiet, pip_args):  # noqa: ARG001
    """🔧 Install package in editable mode."""
    ctx.forward(install, editable=True)


@click.command()
def sdist():
    """📦 Build sdist."""
    run([*PYTHON, "-m", "build", ".", "--sdist"])


@click.command()
def wheel():
    """📦 Build wheel."""
    run([*PYTHON, "-m", "build", ".", "--wheel"])


@click.command()
@click.option("-b", "--builder", default="html", help="Builder to use.")
@click.option("-f", "--force", is_flag=True, help="Ignore cached environment.")
@click.option("-j", "--jobs", default="auto", help="Build in parallel.")
@click.option("-q", "--quiet", is_flag=True, help="Run quietly.")
@click.pass_context
def docs(ctx, builder="html", force=False, jobs="auto", quiet=False):
    """📖 Build Sphinx documentation."""
    sphinx = [*PYTHON, "-m", "sphinx.cmd.build"]
    srcdir = "docs/source"
    outdir = "build"
    options = []
    if quiet:
        options.append("-q")
    if force:
        options.append("-E")
    options.append("-W")
    options.extend(["--jobs", jobs])
    _setup_environment(ctx)
    run([*sphinx, "-M", builder, srcdir, outdir, *options])


@click.command()
@click.pass_context
def browse(ctx):
    """🌐 Browse Sphinx documentation."""
    ctx.invoke(docs, quiet=True)
    browser = [*PYTHON, "-m", "webbrowser", "-n"]
    url = pathlib.Path() / "build" / "html" / "index.html"
    run([*browser, str(url)])



# -----------------------------------------------------------------------------

def _typecode(Datatype datatype: Datatype) -> str | None:
    """
    Map MPI datatype to character code or type string.
    """
    cdef const char *tc = DatatypeCode(datatype.ob_mpi)
    return pystr(tc) if tc != NULL else None


def _typealign(Datatype datatype: Datatype) -> int | None:
    """
    Return MPI datatype alignment.
    """
    cdef size_t align = DatatypeAlign(datatype.ob_mpi)
    return align if align > 0 else None


# -----------------------------------------------------------------------------

cdef inline const char* typechr(const char kind[], size_t size) noexcept nogil:
    cdef char k = kind[0]
    if k == c'b':  # boolean
        if size == 1: return "?"
        if size >= 2: return typechr('i', size)
        return NULL  # ~> unreachable
    if k == c'i':  # signed integral
        if size == sizeof(char)      : return "b"
        if size == sizeof(short)     : return "h"
        if size == sizeof(int)       : return "i"
        if size == sizeof(long)      : return "l"
        if size == sizeof(long long) : return "q"  # ~> uncovered
        return NULL  # ~> unreachable
    if k == c'u':  # unsigned integral
        if size == sizeof(char)      : return "B"
        if size == sizeof(short)     : return "H"
        if size == sizeof(int)       : return "I"
        if size == sizeof(long)      : return "L"
        if size == sizeof(long long) : return "Q"  # ~> uncovered
        return NULL  # ~> unreachable
    if k == c'f':  # real floating
        if size == sizeof(float)//2    : return "e"
        if size == sizeof(float)       : return "f"
        if size == sizeof(double)      : return "d"
        if size == sizeof(long double) : return "g"
        return NULL  # ~> unreachable
    if k == c'c':  # complex floating
        if size == 2*sizeof(float)//2    : return "E"
        if size == 2*sizeof(float)       : return "F"
        if size == 2*sizeof(double)      : return "D"
        if size == 2*sizeof(long double) : return "G"
        return NULL  # ~> unreachable
    return NULL  # ~> unreachable

cdef inline const char* typestr(const char kind[], size_t size) noexcept nogil:
    cdef char k = kind[0]
    if k == c'b':  # boolean
        if size ==  1: return "b1"
        if size >=  2: return typestr('i', size)  # ~> uncovered
        return NULL  # ~> unreachable
    if k == c'i':  # signed integral
        if size ==  1: return "i1"
        if size ==  2: return "i2"
        if size ==  4: return "i4"
        if size ==  8: return "i8"
        if size == 16: return "i16"  # ~> uncovered
        return NULL  # ~> unreachable
    if k == c'u':  # unsigned integral
        if size ==  1: return "u1"
        if size ==  2: return "u2"
        if size ==  4: return "u4"
        if size ==  8: return "u8"
        if size == 16: return "u16"  # ~> uncovered
        return NULL  # ~> unreachable
    if k == c'f':  # real floating
        if size ==  2: return "f2"
        if size ==  4: return "f4"
        if size ==  8: return "f8"
        if size == 12: return "f12"
        if size == 16: return "f16"
        return NULL  # ~> unreachable
    if k == c'c':  # complex floating
        if size ==  4: return "c4"
        if size ==  8: return "c8"
        if size == 16: return "c16"
        if size == 24: return "c24"
        if size == 32: return "c32"
        return NULL  # ~> unreachable
    return NULL  # ~> unreachable

cdef inline const char* typechr_to_typestr(const char tchr[]) noexcept nogil:
    if tchr == NULL: return NULL
    cdef char c = tchr[0]
    # boolean
    if c == c'?': return typestr('b', 1)
    # signed integral
    if c == c'b': return typestr('i', sizeof(char))
    if c == c'h': return typestr('i', sizeof(short))
    if c == c'i': return typestr('i', sizeof(int))
    if c == c'l': return typestr('i', sizeof(long))
    if c == c'q': return typestr('i', sizeof(long long))
    if c == c'p': return typestr('i', sizeof(MPI_Aint))
    # unsigned integral
    if c == c'B': return typestr('u', sizeof(char))
    if c == c'H': return typestr('u', sizeof(short))
    if c == c'I': return typestr('u', sizeof(int))
    if c == c'L': return typestr('u', sizeof(long))
    if c == c'Q': return typestr('u', sizeof(long long))
    if c == c'P': return typestr('u', sizeof(MPI_Aint))
    # floating real
    if c == c'e': return typestr('f', sizeof(float)//2)
    if c == c'f': return typestr('f', sizeof(float))
    if c == c'd': return typestr('f', sizeof(double))
    if c == c'g': return typestr('f', sizeof(long double))
    # floating complex
    if c == c'E': return typestr('c', 2*sizeof(float)//2)
    if c == c'F': return typestr('c', 2*sizeof(float))
    if c == c'D': return typestr('c', 2*sizeof(double))
    if c == c'G': return typestr('c', 2*sizeof(long double))
    # character
    if c == c'S': return "S1"  # NumPy
    if c == c'U': return "U1"  # NumPy
    if c == c'c': return "S1"  # PEP 3118
    if c == c'u': return "u2"  # PEP 3118  # ~> uncovered
    if c == c'w': return "U1"  # PEP 3118  # ~> uncovered
    return NULL  # ~> uncovered

cdef inline const char* mpiaddrchr(size_t size) noexcept nogil:
    if size == sizeof(MPI_Aint)  : return "p"
    if size == sizeof(long long) : return "q"  # ~> uncovered
    if size == sizeof(long)      : return "l"  # ~> uncovered
    if size == sizeof(int)       : return "i"  # ~> uncovered
    return NULL  # ~> uncovered

cdef inline int mpicombiner(MPI_Datatype datatype) noexcept nogil:
    if not mpi_active(): return MPI_COMBINER_NAMED
    cdef int combiner = MPI_COMBINER_NAMED
    cdef MPI_Count ni = 0, na = 0, nc = 0, nd = 0
    <void> MPI_Type_get_envelope_c(datatype, &ni, &na, &nc, &nd, &combiner)
    return combiner

cdef inline size_t mpisizeof(MPI_Datatype datatype) noexcept nogil:
    if not mpi_active(): return 0
    cdef MPI_Count size = MPI_UNDEFINED
    cdef int ierr = MPI_Type_size_c(datatype, &size)
    if ierr != MPI_SUCCESS or size < 0: size = 0
    return <size_t> size

cdef inline MPI_Count mpiextent(MPI_Datatype datatype) noexcept nogil:
    if not mpi_active(): return 0
    cdef MPI_Count lb = 0, extent = MPI_UNDEFINED
    cdef int ierr = MPI_Type_get_extent_c(datatype, &lb, &extent)
    if ierr != MPI_SUCCESS or extent < 0: extent = 0
    return extent

cdef inline const char* mpifortchr(
    const char kind[],
    MPI_Datatype datatype,
) noexcept nogil:
    return typechr(kind, <size_t> mpisizeof(datatype))

cdef inline const char* mpifortstr(
    const char kind[],
    MPI_Datatype datatype,
) noexcept nogil:
    return typestr(kind, <size_t> mpisizeof(datatype))

cdef inline const char* typeDUP(
    const char *(*convert)(MPI_Datatype) noexcept nogil,
    MPI_Datatype datatype,
) noexcept nogil:
    cdef MPI_Datatype basetype = MPI_DATATYPE_NULL
    <void> MPI_Type_get_contents_c(
        datatype, 0, 0, 0, 1, NULL, NULL, NULL, &basetype)
    cdef const char *result = convert(basetype)
    if not predefined(basetype): <void> MPI_Type_free(&basetype)
    return result

cdef extern from * nogil:
    """
    #include <stddef.h>
    #if defined(__cplusplus)
    template<typename T> struct pympi_alignof_struct {char c; T member;};
    #define pympi_alignof(type) offsetof(pympi_alignof_struct<type>, member)
    #else
    #define pympi_alignof(type) offsetof(struct _{char c; type member;}, member)
    #endif
    """
    const size_t alignof_bool       "pympi_alignof(char)"
    const size_t alignof_short      "pympi_alignof(short)"
    const size_t alignof_int        "pympi_alignof(int)"
    const size_t alignof_long       "pympi_alignof(long)"
    const size_t alignof_longlong   "pympi_alignof(long long)"
    const size_t alignof_float      "pympi_alignof(float)"
    const size_t alignof_double     "pympi_alignof(double)"
    const size_t alignof_longdouble "pympi_alignof(long double)"
    const size_t alignof_char       "pympi_alignof(char)"
    const size_t alignof_wchar      "pympi_alignof(wchar_t)"
    const size_t alignof_voidp      "pympi_alignof(void*)"
    const size_t alignof_int128     "16"  # TODO: review
    const size_t alignof_float128   "16"  # TODO: review

cdef inline size_t typealign(const char tchr[]) noexcept nogil:
    if tchr == NULL: return 0
    cdef char c = tchr[0]
    # bool
    if c == c'?': return alignof_bool
    # signed integral
    if c == c'b': return alignof_char
    if c == c'h': return alignof_short
    if c == c'i': return alignof_int
    if c == c'l': return alignof_long
    if c == c'q': return alignof_longlong
    # unsigned integral
    if c == c'B': return alignof_char
    if c == c'H': return alignof_short
    if c == c'I': return alignof_int
    if c == c'L': return alignof_long
    if c == c'Q': return alignof_longlong
    # floating real
    if c == c'e': return alignof_float//2
    if c == c'f': return alignof_float
    if c == c'd': return alignof_double
    if c == c'g': return alignof_longdouble
    # floating complex
    if c == c'E': return alignof_float//2
    if c == c'F': return alignof_float
    if c == c'D': return alignof_double
    if c == c'G': return alignof_longdouble
    # character
    if c == c'c': return alignof_char
    if c == c'S': return alignof_char
    if c == c'U': return alignof_wchar
    # pointer
    if c == c'p': return alignof_voidp
    if c == c'P': return alignof_voidp  # ~> uncovered
    return 0  # ~> unreachable

cdef inline size_t typealignpair(
    const char tc_a[],
    const char tc_b[],
) noexcept nogil:
    cdef size_t align_a = typealign(tc_a)
    cdef size_t align_b = typealign(tc_b)
    return align_a if align_a > align_b else align_b

# -----------------------------------------------------------------------------

cdef inline const char* DatatypeChar(MPI_Datatype datatype) noexcept nogil:
    if datatype == MPI_DATATYPE_NULL: return NULL
    # MPI
    if datatype == MPI_PACKED : return "B"
    if datatype == MPI_BYTE   : return "B"
    if datatype == MPI_AINT   : return mpiaddrchr(sizeof(MPI_Aint))
    if datatype == MPI_OFFSET : return mpiaddrchr(sizeof(MPI_Offset))
    if datatype == MPI_COUNT  : return mpiaddrchr(sizeof(MPI_Count))
    # C - character
    if datatype == MPI_CHAR : return "c"
    if datatype == MPI_WCHAR and sizeof(wchar_t) == 2: return typechr('u', 2)
    if datatype == MPI_WCHAR and sizeof(wchar_t) == 4: return "U"
    # C - (signed) integral
    if datatype == MPI_SIGNED_CHAR : return "b"
    if datatype == MPI_SHORT       : return "h"
    if datatype == MPI_INT         : return "i"
    if datatype == MPI_LONG        : return "l"
    if datatype == MPI_LONG_LONG   : return "q"
    # C - unsigned integral
    if datatype == MPI_UNSIGNED_CHAR      : return "B"
    if datatype == MPI_UNSIGNED_SHORT     : return "H"
    if datatype == MPI_UNSIGNED           : return "I"
    if datatype == MPI_UNSIGNED_LONG      : return "L"
    if datatype == MPI_UNSIGNED_LONG_LONG : return "Q"
    # C - (real) floating
    if datatype == MPI_FLOAT       : return "f"
    if datatype == MPI_DOUBLE      : return "d"
    if datatype == MPI_LONG_DOUBLE : return "g"
    # C99 - boolean
    if datatype == MPI_C_BOOL : return "?"
    # C99 - integral
    if datatype == MPI_INT8_T   : return typechr('i', 1)
    if datatype == MPI_INT16_T  : return typechr('i', 2)
    if datatype == MPI_INT32_T  : return typechr('i', 4)
    if datatype == MPI_INT64_T  : return typechr('i', 8)
    if datatype == MPI_UINT8_T  : return typechr('u', 1)
    if datatype == MPI_UINT16_T : return typechr('u', 2)
    if datatype == MPI_UINT32_T : return typechr('u', 4)
    if datatype == MPI_UINT64_T : return typechr('u', 8)
    # C99 - real/complex floating
    if datatype == MPI_FLOAT16_T             : return "e"
    if datatype == MPI_C_COMPLEX             : return "F"
    if datatype == MPI_C_FLOAT_COMPLEX       : return "F"
    if datatype == MPI_C_DOUBLE_COMPLEX      : return "D"
    if datatype == MPI_C_LONG_DOUBLE_COMPLEX : return "G"
    # C++ - boolean
    if datatype == MPI_CXX_BOOL : return "?"
    # C++ - complex floating
    if datatype == MPI_CXX_FLOAT_COMPLEX       : return "F"
    if datatype == MPI_CXX_DOUBLE_COMPLEX      : return "D"
    if datatype == MPI_CXX_LONG_DOUBLE_COMPLEX : return "G"
    # Fortran 77
    if datatype == MPI_CHARACTER        : return "c"
    if datatype == MPI_LOGICAL          : return mpifortchr('b', datatype)
    if datatype == MPI_INTEGER          : return mpifortchr('i', datatype)
    if datatype == MPI_REAL             : return mpifortchr('f', datatype)
    if datatype == MPI_DOUBLE_PRECISION : return mpifortchr('f', datatype)
    if datatype == MPI_COMPLEX          : return mpifortchr('c', datatype)
    if datatype == MPI_DOUBLE_COMPLEX   : return mpifortchr('c', datatype)
    # Fortran 90
    if datatype == MPI_LOGICAL1  : return typechr('b',  1)
    if datatype == MPI_LOGICAL2  : return typechr('b',  2)
    if datatype == MPI_LOGICAL4  : return typechr('b',  4)
    if datatype == MPI_LOGICAL8  : return typechr('b',  8)
    if datatype == MPI_LOGICAL16 : return typechr('b', 16)
    if datatype == MPI_INTEGER1  : return typechr('i',  1)
    if datatype == MPI_INTEGER2  : return typechr('i',  2)
    if datatype == MPI_INTEGER4  : return typechr('i',  4)
    if datatype == MPI_INTEGER8  : return typechr('i',  8)
    if datatype == MPI_INTEGER16 : return typechr('i', 16)
    if datatype == MPI_REAL2     : return typechr('f',  2)
    if datatype == MPI_REAL4     : return typechr('f',  4)
    if datatype == MPI_REAL8     : return typechr('f',  8)
    if datatype == MPI_REAL16    : return typechr('f', 16)
    if datatype == MPI_COMPLEX4  : return typechr('c',  4)
    if datatype == MPI_COMPLEX8  : return typechr('c',  8)
    if datatype == MPI_COMPLEX16 : return typechr('c', 16)
    if datatype == MPI_COMPLEX32 : return typechr('c', 32)
    cdef int combiner = mpicombiner(datatype)
    if combiner == MPI_COMBINER_F90_INTEGER : return mpifortchr('i', datatype)
    if combiner == MPI_COMBINER_F90_REAL    : return mpifortchr('f', datatype)
    if combiner == MPI_COMBINER_F90_COMPLEX : return mpifortchr('c', datatype)
    # duplicate
    if combiner == MPI_COMBINER_DUP: return typeDUP(DatatypeChar, datatype)
    # fallback
    return NULL

# -----------------------------------------------------------------------------

cdef inline const char* DatatypeStr(MPI_Datatype datatype) noexcept nogil:
    if datatype == MPI_DATATYPE_NULL : return NULL
    # C99
    if datatype == MPI_C_BOOL   : return typestr('b', 1)
    if datatype == MPI_INT8_T   : return typestr('i', 1)
    if datatype == MPI_INT16_T  : return typestr('i', 2)
    if datatype == MPI_INT32_T  : return typestr('i', 4)
    if datatype == MPI_INT64_T  : return typestr('i', 8)
    if datatype == MPI_UINT8_T  : return typestr('u', 1)
    if datatype == MPI_UINT16_T : return typestr('u', 2)
    if datatype == MPI_UINT32_T : return typestr('u', 4)
    if datatype == MPI_UINT64_T : return typestr('u', 8)
    # Fortran 90
    if datatype == MPI_LOGICAL1  : return typestr('b',  1)
    if datatype == MPI_LOGICAL2  : return typestr('b',  2)
    if datatype == MPI_LOGICAL4  : return typestr('b',  4)
    if datatype == MPI_LOGICAL8  : return typestr('b',  8)
    if datatype == MPI_LOGICAL16 : return typestr('b', 16)
    if datatype == MPI_INTEGER1  : return typestr('i',  1)
    if datatype == MPI_INTEGER2  : return typestr('i',  2)
    if datatype == MPI_INTEGER4  : return typestr('i',  4)
    if datatype == MPI_INTEGER8  : return typestr('i',  8)
    if datatype == MPI_INTEGER16 : return typestr('i', 16)
    if datatype == MPI_REAL2     : return typestr('f',  2)
    if datatype == MPI_REAL4     : return typestr('f',  4)
    if datatype == MPI_REAL8     : return typestr('f',  8)
    if datatype == MPI_REAL16    : return typestr('f', 16)
    if datatype == MPI_COMPLEX4  : return typestr('c',  4)
    if datatype == MPI_COMPLEX8  : return typestr('c',  8)
    if datatype == MPI_COMPLEX16 : return typestr('c', 16)
    if datatype == MPI_COMPLEX32 : return typestr('c', 32)
    cdef int combiner = mpicombiner(datatype)
    if combiner == MPI_COMBINER_F90_INTEGER : return mpifortstr('i', datatype)
    if combiner == MPI_COMBINER_F90_REAL    : return mpifortstr('f', datatype)
    if combiner == MPI_COMBINER_F90_COMPLEX : return mpifortstr('c', datatype)
    # duplicate
    if combiner == MPI_COMBINER_DUP: return typeDUP(DatatypeStr, datatype)
    # fallback
    return typechr_to_typestr(DatatypeChar(datatype))

# -----------------------------------------------------------------------------

cdef inline const char* DatatypeCode(MPI_Datatype datatype) noexcept nogil:
    if datatype == MPI_DATATYPE_NULL : return NULL
    # C99
    if datatype == MPI_C_BOOL   : return typestr('b', 1)
    if datatype == MPI_INT8_T   : return typestr('i', 1)
    if datatype == MPI_INT16_T  : return typestr('i', 2)
    if datatype == MPI_INT32_T  : return typestr('i', 4)
    if datatype == MPI_INT64_T  : return typestr('i', 8)
    if datatype == MPI_UINT8_T  : return typestr('u', 1)
    if datatype == MPI_UINT16_T : return typestr('u', 2)
    if datatype == MPI_UINT32_T : return typestr('u', 4)
    if datatype == MPI_UINT64_T : return typestr('u', 8)
    # Fortran 90
    if datatype == MPI_LOGICAL1  : return typestr('b',  1)
    if datatype == MPI_LOGICAL2  : return typestr('b',  2)
    if datatype == MPI_LOGICAL4  : return typestr('b',  4)
    if datatype == MPI_LOGICAL8  : return typestr('b',  8)
    if datatype == MPI_LOGICAL16 : return typestr('b', 16)
    if datatype == MPI_INTEGER1  : return typestr('i',  1)
    if datatype == MPI_INTEGER2  : return typestr('i',  2)
    if datatype == MPI_INTEGER4  : return typestr('i',  4)
    if datatype == MPI_INTEGER8  : return typestr('i',  8)
    if datatype == MPI_INTEGER16 : return typestr('i', 16)
    if datatype == MPI_REAL2     : return typestr('f',  2)
    if datatype == MPI_REAL4     : return typestr('f',  4)
    if datatype == MPI_REAL8     : return typestr('f',  8)
    if datatype == MPI_REAL16    : return typestr('f', 16)
    if datatype == MPI_COMPLEX4  : return typestr('c',  4)
    if datatype == MPI_COMPLEX8  : return typestr('c',  8)
    if datatype == MPI_COMPLEX16 : return typestr('c', 16)
    if datatype == MPI_COMPLEX32 : return typestr('c', 32)
    cdef int combiner = mpicombiner(datatype)
    if combiner == MPI_COMBINER_F90_INTEGER : return mpifortstr('i', datatype)
    if combiner == MPI_COMBINER_F90_REAL    : return mpifortstr('f', datatype)
    if combiner == MPI_COMBINER_F90_COMPLEX : return mpifortstr('c', datatype)
    # duplicate
    if combiner == MPI_COMBINER_DUP: return typeDUP(DatatypeCode, datatype)
    # fallback
    return DatatypeChar(datatype)

# -----------------------------------------------------------------------------

cdef inline size_t DatatypeAlign(MPI_Datatype datatype) noexcept nogil:
    if datatype == MPI_DATATYPE_NULL   : return 0
    if datatype == MPI_LOGICAL16       : return alignof_int128
    if datatype == MPI_INTEGER16       : return alignof_int128
    if datatype == MPI_REAL16          : return alignof_float128
    if datatype == MPI_COMPLEX32       : return alignof_float128
    cdef size_t align = typealign(DatatypeChar(datatype))
    if align > 0: return align
    if datatype == MPI_SHORT_INT       : return typealignpair("h", "i")
    if datatype == MPI_2INT            : return typealignpair("i", "i")
    if datatype == MPI_LONG_INT        : return typealignpair("l", "i")
    if datatype == MPI_FLOAT_INT       : return typealignpair("f", "i")
    if datatype == MPI_DOUBLE_INT      : return typealignpair("d", "i")
    if datatype == MPI_LONG_DOUBLE_INT : return typealignpair("g", "i")
    return 0

# -----------------------------------------------------------------------------



# -----------------------------------------------------------------------------


@cython.final
cdef class BottomType:
    """
    Type of `BOTTOM`.
    """

    cdef inline str as_str(self):
        return 'BOTTOM'
    cdef inline void* as_buf(self) noexcept nogil:
        return MPI_BOTTOM
    cdef inline MPI_Aint as_int(self) noexcept nogil:
        return <MPI_Aint>MPI_BOTTOM

    def __bool__(self) -> bool:
        return self.as_int() != 0

    def __int__(self) -> int:
        return self.as_int()

    def __index__(self) -> int:
        return self.as_int()

    def __hash__(self) -> int:
        return <Py_hash_t>self.as_int()

    def __eq__(self, object value) -> bool:
        if isinstance(value, type(self)): return True
        if not is_integral(value): return NotImplemented
        return self.as_int() == <MPI_Aint>value

    def __ne__(self, object value) -> bool:
        if isinstance(value, type(self)): return False
        if not is_integral(value): return NotImplemented
        return self.as_int() != <MPI_Aint>value

    def __getbuffer__(self, Py_buffer *view, int flags):
        PyBuffer_FillInfo(view, <object>NULL, self.as_buf(), 0, 0, flags)

    def __repr__(self) -> str:
        return self.as_str()

    def __reduce__(self) -> str:
        return self.as_str()


cdef object __BOTTOM__ = BottomType()

cdef inline bint is_BOTTOM(object obj):
    return obj is None or is_constant(obj, __BOTTOM__)

# -----------------------------------------------------------------------------


@cython.final
cdef class InPlaceType:
    """
    Type of `IN_PLACE`.
    """

    cdef inline str as_str(self):
        return 'IN_PLACE'
    cdef inline void* as_buf(self) noexcept nogil:
        return MPI_IN_PLACE
    cdef inline MPI_Aint as_int(self) noexcept nogil:
        return <MPI_Aint>MPI_IN_PLACE

    def __bool__(self) -> bool:
        return self.as_int() != 0

    def __int__(self) -> int:
        return self.as_int()

    def __hash__(self) -> int:
        return <Py_hash_t>self.as_int()

    def __eq__(self, object value) -> bool:
        if isinstance(value, type(self)): return True
        if not is_integral(value): return NotImplemented
        return self.as_int() == <MPI_Aint>value

    def __ne__(self, object value) -> bool:
        if isinstance(value, type(self)): return False
        if not is_integral(value): return NotImplemented
        return self.as_int() != <MPI_Aint>value

    def __getbuffer__(self, Py_buffer *view, int flags):
        PyBuffer_FillInfo(view, <object>NULL, self.as_buf(), 0, 0, flags)

    def __repr__(self) -> str:
        return self.as_str()

    def __reduce__(self) -> str:
        return self.as_str()


cdef object __IN_PLACE__ = InPlaceType()

cdef inline bint is_IN_PLACE(object obj):
    return obj is None or is_constant(obj, __IN_PLACE__)

# -----------------------------------------------------------------------------


@cython.final
cdef class BufferAutomaticType:
    """
    Type of `BUFFER_AUTOMATIC`.
    """

    cdef inline object as_str(self):
        return 'BUFFER_AUTOMATIC'
    cdef inline void* as_buf(self) noexcept nogil:
        return MPI_BUFFER_AUTOMATIC
    cdef inline MPI_Aint as_int(self) noexcept nogil:
        return <MPI_Aint>MPI_BUFFER_AUTOMATIC

    def __bool__(self) -> bool:
        return self.as_int() != 0

    def __int__(self) -> int:
        return self.as_int()

    def __hash__(self) -> int:
        return <Py_hash_t>self.as_int()

    def __eq__(self, object value) -> bool:
        if isinstance(value, type(self)): return True
        if not is_integral(value): return NotImplemented
        return self.as_int() == <MPI_Aint>value

    def __ne__(self, object value) -> bool:
        if isinstance(value, type(self)): return False
        if not is_integral(value): return NotImplemented
        return self.as_int() != <MPI_Aint>value

    def __getbuffer__(self, Py_buffer *view, int flags):
        PyBuffer_FillInfo(view, <object>NULL, self.as_buf(), 0, 0, flags)

    def __repr__(self) -> str:
        return self.as_str()

    def __reduce__(self) -> str:
        return self.as_str()


cdef object __BUFFER_AUTOMATIC__ = BufferAutomaticType()

cdef inline bint is_BUFFER_AUTOMATIC(object obj):
    return is_constant(obj, __BUFFER_AUTOMATIC__)

# -----------------------------------------------------------------------------



import importlib
import sys

sys.path.insert(0, sys.argv[1])
MPI = importlib.import_module("mpi4py.MPI")

parent = MPI.Comm.Get_parent()
parent.Barrier()
parent.Disconnect()
assert parent == MPI.COMM_NULL
parent = MPI.Comm.Get_parent()
assert parent == MPI.COMM_NULL



import concurrent.futures as cf
import functools
import itertools
import os
import pathlib
import platform
import sys
import time
import unittest
import warnings

import mpi4py.util.pool as pool
from mpi4py import MPI

try:
    import mpitestutil as testutil
except ImportError:
    sys.path.append(os.fspath(pathlib.Path(__file__).resolve().parent))
    import mpitestutil as testutil


mpi_vendor_name, mpi_vendor_version = MPI.get_vendor()


def sqr(x, wait=0.0):
    time.sleep(wait)
    return x * x


def mul(x, y):  # noqa: FURB118
    return x * y


def identity(x):
    return x


def raising():
    raise KeyError("key")


TIMEOUT1 = 0.1
TIMEOUT2 = 0.2


class TimingWrapper:
    #
    def __init__(self, func):
        self.func = func
        self.elapsed = None

    def __call__(self, *args, **kwds):
        t = time.monotonic()
        try:
            return self.func(*args, **kwds)
        finally:
            self.elapsed = time.monotonic() - t


class BaseTestPool:
    #
    PoolType = None

    @classmethod
    def Pool(cls, *args, **kwargs):
        if "coverage" in sys.modules:
            kwargs["python_args"] = "-m coverage run".split()
        Pool = cls.PoolType
        return Pool(*args, **kwargs)

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.pool = cls.Pool(1)

    @classmethod
    def tearDownClass(cls):
        cls.pool.terminate()
        cls.pool.join()
        cls.pool = None
        super().tearDownClass()

    @unittest.skipIf(
        mpi_vendor_name == "Open MPI"
        and mpi_vendor_version < (5, 0, 0)
        and platform.machine() == "s390x",
        "openmpi(<5.0.0)-s390x",
    )
    def test_apply(self):
        papply = self.pool.apply
        self.assertEqual(papply(sqr, (5,)), sqr(5))
        self.assertEqual(papply(sqr, (), {"x": 3}), sqr(x=3))

    def test_map(self):
        self.assertEqual(
            self.pool.map(sqr, range(10)), list(map(sqr, list(range(10))))
        )
        self.assertEqual(
            self.pool.map(sqr, (i for i in range(10))),
            list(map(sqr, list(range(10)))),
        )
        self.assertEqual(
            self.pool.map(sqr, list(range(10))),
            list(map(sqr, list(range(10)))),
        )

        self.assertEqual(
            self.pool.map(sqr, range(100), chunksize=20),
            list(map(sqr, list(range(100)))),
        )
        self.assertEqual(
            self.pool.map(sqr, (i for i in range(100)), chunksize=20),
            list(map(sqr, list(range(100)))),
        )
        self.assertEqual(
            self.pool.map(sqr, list(range(100)), chunksize=20),
            list(map(sqr, list(range(100)))),
        )

    def test_imap(self):
        self.assertEqual(
            list(self.pool.imap(sqr, range(10))),
            list(map(sqr, list(range(10)))),
        )
        self.assertEqual(
            list(self.pool.imap(sqr, (i for i in range(10)))),
            list(map(sqr, list(range(10)))),
        )
        self.assertEqual(
            list(self.pool.imap(sqr, list(range(10)))),
            list(map(sqr, list(range(10)))),
        )

        it = self.pool.imap(sqr, range(10))
        for i in range(10):
            self.assertEqual(next(it), i * i)
        self.assertRaises(StopIteration, next, it)
        it = self.pool.imap(sqr, list(range(10)))
        for i in range(10):
            self.assertEqual(next(it), i * i)
        self.assertRaises(StopIteration, next, it)

        it = self.pool.imap(sqr, range(100), chunksize=20)
        for i in range(100):
            self.assertEqual(next(it), i * i)
        self.assertRaises(StopIteration, next, it)
        it = self.pool.imap(sqr, list(range(100)), chunksize=20)
        for i in range(100):
            self.assertEqual(next(it), i * i)
        self.assertRaises(StopIteration, next, it)

    def test_imap_unordered(self):
        args = list(range(10))
        result = list(map(sqr, args))
        it = self.pool.imap_unordered(sqr, args)
        self.assertEqual(sorted(it), result)
        it = self.pool.imap_unordered(sqr, iter(args))
        self.assertEqual(sorted(it), result)
        it = self.pool.imap_unordered(sqr, (a for a in args))
        self.assertEqual(sorted(it), result)

        args = list(range(100))
        result = list(map(sqr, args))
        it = self.pool.imap_unordered(sqr, args, chunksize=20)
        self.assertEqual(sorted(it), result)
        it = self.pool.imap_unordered(sqr, iter(args), chunksize=20)
        self.assertEqual(sorted(it), result)
        it = self.pool.imap_unordered(sqr, (a for a in args), chunksize=20)
        self.assertEqual(sorted(it), result)

    def test_starmap(self):
        tuples = list(zip(range(10), range(9, -1, -1)))
        self.assertEqual(
            self.pool.starmap(mul, tuples),
            list(itertools.starmap(mul, tuples)),
        )
        tuples = list(zip(range(100), range(99, -1, -1)))
        self.assertEqual(
            self.pool.starmap(mul, tuples, chunksize=20),
            list(itertools.starmap(mul, tuples)),
        )

    def test_istarmap(self):
        tuples = list(zip(range(10), range(9, -1, -1)))
        result = list(itertools.starmap(mul, tuples))
        it = self.pool.istarmap(mul, tuples)
        self.assertEqual(list(it), result)
        iterator = zip(range(10), range(9, -1, -1))
        it = self.pool.istarmap(mul, iterator)
        self.assertEqual(list(it), result)

        tuples = list(zip(range(10), range(9, -1, -1)))
        it = self.pool.istarmap(mul, tuples)
        for i, j in tuples:
            self.assertEqual(next(it), i * j)
        self.assertRaises(StopIteration, next, it)

        tuples = list(zip(range(100), range(99, -1, -1)))
        it = self.pool.istarmap(mul, tuples, chunksize=20)
        for i, j in tuples:
            self.assertEqual(next(it), i * j)
        self.assertRaises(StopIteration, next, it)

    def test_istarmap_unordered(self):
        tuples = list(zip(range(10), range(9, -1, -1)))
        result = list(itertools.starmap(mul, tuples))
        it = self.pool.istarmap_unordered(mul, tuples)
        self.assertEqual(sorted(it), sorted(result))
        iterator = zip(range(10), range(9, -1, -1))
        it = self.pool.istarmap_unordered(mul, iterator)
        self.assertEqual(sorted(it), sorted(result))

        tuples = list(zip(range(100), range(99, -1, -1)))
        result = list(itertools.starmap(mul, tuples))
        it = self.pool.istarmap_unordered(mul, tuples, chunksize=20)
        self.assertEqual(sorted(it), sorted(result))

    def test_apply_async(self):
        res = self.pool.apply_async(sqr, (7,))
        self.assertEqual(res.get(), 49)

        res = self.pool.apply_async(sqr, (7, TIMEOUT2))
        get = TimingWrapper(res.get)
        self.assertEqual(get(), 49)
        self.assertLess(get.elapsed, TIMEOUT2 * 10)
        self.assertGreater(get.elapsed, TIMEOUT2 / 10)

    @unittest.skipIf(
        testutil.github()
        and sys.platform == "darwin"
        and sys.implementation.name == "cpython"
        and sys.abiflags == "t",
        "gha-macos-freethreaded",
    )
    def test_apply_async_timeout(self):
        res = self.pool.apply_async(sqr, (7, TIMEOUT2))
        self.assertFalse(res.ready())
        self.assertRaises(ValueError, res.successful)
        res.wait(TIMEOUT2 / 100)
        self.assertFalse(res.ready())
        self.assertRaises(ValueError, res.successful)
        self.assertRaises(TimeoutError, res.get, TIMEOUT2 / 100)
        res.wait()
        self.assertTrue(res.ready())
        self.assertTrue(res.successful())
        self.assertEqual(res.get(), 49)

    def test_map_async(self):
        args = list(range(10))
        self.assertEqual(
            self.pool.map_async(sqr, args).get(), list(map(sqr, args))
        )
        args = list(range(100))
        self.assertEqual(
            self.pool.map_async(sqr, args, chunksize=20).get(),
            list(map(sqr, args)),
        )

    def test_map_async_callbacks(self):
        call_args = []
        result = self.pool.map_async(
            int,
            ["1", "2"],
            callback=call_args.append,
            error_callback=call_args.append,
        )
        result.wait()
        self.assertTrue(result.successful())
        self.assertEqual(len(call_args), 1)
        self.assertEqual(call_args[0], [1, 2])
        result = self.pool.map_async(
            int,
            ["a"],
            callback=call_args.append,
            error_callback=call_args.append,
        )
        result.wait()
        self.assertFalse(result.successful())
        self.assertEqual(len(call_args), 2)
        self.assertIsInstance(call_args[1], ValueError)

    def test_starmap_async(self):
        tuples = list(zip(range(10), range(9, -1, -1)))
        self.assertEqual(
            self.pool.starmap_async(mul, tuples).get(),
            list(itertools.starmap(mul, tuples)),
        )
        tuples = list(zip(range(1000), range(999, -1, -1)))
        self.assertEqual(
            self.pool.starmap_async(mul, tuples, chunksize=100).get(),
            list(itertools.starmap(mul, tuples)),
        )

    # ---

    def test_terminate(self):
        p = self.Pool(1)
        for _ in range(100):
            p.apply_async(time.sleep, (TIMEOUT1,))
        result = p.apply_async(time.sleep, (TIMEOUT1,))
        p.terminate()
        p.join()
        result.wait(TIMEOUT1)
        result.wait()
        self.assertFalse(result.successful())
        self.assertRaises(TimeoutError, result.get, TIMEOUT1)
        self.assertRaises(TimeoutError, result.get)

        p = self.Pool(1)
        args = [TIMEOUT1] * 100
        result = p.map_async(time.sleep, args, chunksize=1)
        p.terminate()
        p.join()
        result.wait(TIMEOUT1)
        result.wait()
        self.assertFalse(result.successful())
        self.assertRaises(TimeoutError, result.get, TIMEOUT1)
        self.assertRaises(TimeoutError, result.get)

    def test_empty_iterable(self):
        p = self.Pool(1)

        self.assertEqual(p.map(sqr, []), [])
        self.assertEqual(list(p.imap(sqr, [])), [])
        self.assertEqual(list(p.imap_unordered(sqr, [])), [])

        self.assertEqual(p.starmap(sqr, []), [])
        self.assertEqual(list(p.istarmap(sqr, [])), [])
        self.assertEqual(list(p.istarmap_unordered(sqr, [])), [])

        self.assertEqual(p.map_async(sqr, []).get(), [])
        self.assertEqual(p.starmap_async(mul, []).get(), [])

        p.close()
        p.join()

    def test_enter_exit(self):
        pool = self.Pool(1)

        with pool:
            pass
        # with self.assertRaises(ValueError):
        #     with pool:
        #         pass

        pool.join()

    # ---

    def test_async_error_callback(self):
        p = self.Pool(1)
        scratchpad = [None]

        def errback(exc):
            scratchpad[0] = exc

        res = p.apply_async(raising, error_callback=errback)
        p.close()
        p.join()
        self.assertRaises(KeyError, res.get)
        self.assertTrue(scratchpad[0])
        self.assertIsInstance(scratchpad[0], KeyError)

    def test_pool_worker_lifetime_early_close(self):
        p = self.Pool(1)
        results = []
        for i in range(5):
            results.append(p.apply_async(sqr, (i, TIMEOUT1)))
        p.close()
        p.join()
        for j, res in enumerate(results):
            self.assertEqual(res.get(), sqr(j))

    # ---

    def test_arg_processes(self):
        with self.assertRaises(ValueError):
            self.Pool(-1)
        with self.assertRaises(ValueError):
            self.Pool(0)

    def test_arg_initializer(self):
        self.Pool(1, initializer=identity, initargs=(123,))
        with self.assertRaises(TypeError):
            self.Pool(initializer=123)

    def test_unsupported_args(self):
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            with self.assertRaises(UserWarning):
                with self.Pool(1, maxtasksperchild=1):
                    pass
            with self.assertRaises(UserWarning):
                with self.Pool(1, context=object):
                    pass


# ---


@unittest.skipIf(testutil.disable_mpi_spawn(), "mpi-spawn")
@unittest.skipIf(MPI.COMM_WORLD.Get_size() > 1, "mpi-world-size>1")
class TestProcessPool(BaseTestPool, unittest.TestCase):
    #
    PoolType = pool.Pool


class TestThreadPool(BaseTestPool, unittest.TestCase):
    #
    PoolType = pool.ThreadPool


# ---


class ExtraExecutorMixing:
    #
    def map(
        self,
        fn,
        iterable,
        timeout=None,
        chunksize=1,
        unordered=False,
    ):
        del unordered  # ignored, unused
        return super().map(
            fn,
            iterable,
            timeout=timeout,
            chunksize=chunksize,
        )

    def starmap(
        self,
        fn,
        iterable,
        timeout=None,
        chunksize=1,
        unordered=False,
    ):
        del unordered  # ignored, unused
        fn = functools.partial(self._apply_args, fn)
        return super().map(
            fn,
            iterable,
            timeout=timeout,
            chunksize=chunksize,
        )

    @staticmethod
    def _apply_args(fn, args):
        return fn(*args)


class ExtraExecutor(ExtraExecutorMixing, cf.ThreadPoolExecutor):
    #
    pass


class ExtraPool(pool.Pool):
    #
    Executor = ExtraExecutor


class TestExtraPool(BaseTestPool, unittest.TestCase):
    #
    PoolType = ExtraPool

    @classmethod
    def Pool(cls, *args, **kwargs):
        return cls.PoolType(*args, **kwargs)


del TestExtraPool


# ---

if __name__ == "__main__":
    unittest.main()
